# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

"""
This is an old module which is deprecated now. It is kept in the project as it can be used to read BIF6 and BIF3D images.
"""

import os
import re
import struct

import matplotlib.pyplot as plt
import numpy as np

import pySPM

Elements = {
    "H": {1: (1.00782503223, 0.999885), 2: (2.01410177812, 0.000115)},
    "Li": {6: (6.0151228874, 0.0759), 7: (7.0160034366, 0.9241)},
    "Be": {9: (9.012183065, 1)},
    "B": {10: (10.01293695, 0.199), 11: (11.00930536, 0.801)},
    "C": {12: (12, 0.9893), 13: (13.00335483507, 0.01107)},
    "N": {14: (14.00307400443, 0.99636), 15: (15.00010889888, 0.00364)},
    "O": {
        16: (15.99491461957, 0.99757),
        17: (16.99913175650, 0.00038),
        18: (17.99915961286, 0.00205),
    },
    "F": {19: (18.9984036273, 1)},
    "Ag": {107: (106.9050916, 0.51839), 109: (108.9047553, 0.48161)},
    "Au": {197: (196.96656879, 1)},
    "Si": {
        28: (27.97692653465, 0.92223),
        29: (28.97649466490, 0.04685),
        30: (29.973770136, 0.3092),
    },
    "Ti": {
        46: (45.95262772, 0.0825),
        47: (46.95175879, 0.0744),
        48: (47.94794198, 0.7372),
        49: (48.94786568, 0.0541),
        50: (49.94478689, 0.0518),
    },
}


def getSpecElt(Elts):
    if len(Elts) == 1:
        return Elements[Elts[0]]
    tm = {}
    r1 = Elements[Elts[0]]
    r2 = getSpecElt(Elts[1:])
    for x in r1:
        for y in r2:
            if x + y not in tm:
                tm[x + y] = [r1[x][0] + r2[y][0], r1[x][1] * r2[y][1]]
            else:
                tm[x + y][1] += r1[x][1] * r2[y][1]
    return tm


def SplitElts(elt):
    elts = []
    for x, i in re.findall("([A-Z][a-z]?)([0-9]*)", elt):
        i = 1 if i == "" else int(i)
        elts += [x for k in range(i)]
    return elts


def showElts(Elts):
    r = getSpecElt(Elts)
    plt.bar(r.keys(), [z[1] for z in r.values()])
    plt.show()


class BIF6:
    def __init__(self, filename):
        self.f = open(filename, "rb")
        self.header = struct.unpack("xx4s5H", self.f.read(16))
        self.size = (self.header[2], self.header[3])
        self.N = self.size[0] * self.size[1]
        self.cat = []
        for i in range(self.header[1]):
            self.f.seek(16 + i * (4 * self.N + 16))
            self.cat.append(struct.unpack("4f", self.f.read(16)))

    def getImgID(self, ID):
        assert ID >= 0 and self.header[1] >= ID
        self.f.seek(32 + ID * (4 * self.N + 16))
        return np.array(
            struct.unpack(str(self.N) + "I", self.f.read(4 * self.N))
        ).reshape(self.size)

    def getImgMass(self, masses, raw=False):
        if isinstance(masses, (float, int)):
            masses = [masses]
        SUM = None
        for i, x in enumerate(self.cat):
            for m in masses:
                if m >= x[0] - 0.5 and m <= x[1] + 0.5:
                    if SUM is None:
                        SUM = self.getImgID(i)
                    else:
                        SUM += self.getImgID(i)
        if raw:
            return SUM
        if SUM is None:
            return None
        return pySPM.SPM_image(np.flipud(SUM))

    def getImgElt(self, elt):
        r = {}
        A = getSpecElt(SplitElts(elt))
        for k in A:
            I = self.getImgMass(k)
            if I is not None:
                r[k] = {"data": I, "mass": A[k][0], "abund": A[k][1]}
        return r

    def showImgElt(self, elt, size=10, abundCorr=False, tot=True):
        A = self.getImgElt(elt)
        ks = [z for z in A if A[z]["data"] is not None]
        fig, ax = plt.subplots(
            (len(ks) + 1) // 4 + 1, 4, figsize=(10 * ((len(ks) + 1) // 4 + 1), 10)
        )
        mi = np.min(A[ks[0]]["data"].pixels)
        ma = np.max(A[ks[0]]["data"].pixels)
        A[ks[0]]["CT"] = ma
        for _i, k in enumerate(ks[1:]):
            M = np.min(A[k]["data"].pixels)
            mi = min(mi, M)
            A[k]["minCT"] = M
            M = np.max(A[k]["data"].pixels)
            ma = max(ma, M)
            A[k]["CT"] = M
        ks.sort()
        di = 0
        if tot:
            SUM = sum([A[k]["data"].pixels for k in ks])
            ax[0][0].imshow(SUM, vmin=0)
            ax[0][0].set_title("Total")
            di = 1
        for _i, k in enumerate(ks):
            if abundCorr:
                ax[(_i + di) // 4][(_i + di) % 4].imshow(
                    A[k]["data"].pixels / A[k]["abund"], vmin=0, vmax=ma
                )
            else:
                A[k]["data"].show(ax=ax[(_i + di) // 4][(_i + di) % 4], vmin=0, vmax=ma)
            ax[(_i + di) // 4][(_i + di) % 4].set_title(
                "mass: {mass:.3} - abundancy: {abund:.3f} - CT: {CT}".format(**A[k])
            )

    def __exit__(self, exc_type, exc_value, traceback):
        self.f.close()


class BIF3D:
    def __init__(self, path):
        self.Peaks = {}
        self.RPeaks = {}
        self.Path = os.path.dirname(path)
        self.Basename = os.path.basename(path)
        self.size = None
        for x in os.listdir(self.Path):
            if x[: len(self.Basename)] == self.Basename:
                if x[-6:] == ".BIF3D":
                    s = self.Basename + r" \(([0-9]+)\)(?: - (.*?))?\.BIF3D"
                    r = re.search(s, x)
                    ID = int(r.group(1))
                    self.Peaks[ID] = r.group(2)
                    if r.group(2) is not None:
                        self.RPeaks[r.group(2)] = ID
        self.data = {}

    def getIDs(self, channels):
        assert type(channels) in [str, int, list, tuple]
        if isinstance(channels, int):
            return channels
        if isinstance(channels, str):
            assert channels in self.RPeaks
            return self.RPeaks[channels]
        if isinstance(channels, (list, tuple)):
            ID = []
            for x in channels:
                ID.append(self.getIDs(x))
            return ID

    def getChannel(self, channel):
        k = self.getIDs(channel)
        if k in self.data:
            return self.data[k]
        v = self.Peaks[k]
        if v is not None:
            path = "{path}{sep}{basename} ({ID}) - {Peak}.BIF3D"
        else:
            path = "{path}{sep}{basename} ({ID}).BIF3D"
        f = open(
            path.format(
                path=self.Path, sep=os.sep, basename=self.Basename, ID=k, Peak=v
            ),
            "rb",
        )
        A = f.read()
        f.close()
        size = struct.unpack("II", A[32:40])
        if self.size is None:
            self.size = size
        else:
            assert self.size == size
        return np.array(
            struct.unpack(f"{self.size[0] * self.size[1]}d", A[640:])
        ).reshape(self.size)

    def loadChannel(self, channel):
        k = self.getIDs(channel)
        if k not in self.data:
            self.data[k] = self.getChannel(channel)

    def loadChannels(self, channels):
        if type(channels) not in [list, tuple]:
            self.loadChannel(channels)
        else:
            for x in channels:
                self.loadChannel(x)

    def getChannels(self, *channels):
        k = self.getIDs(channels)
        self.loadChannels(k)
        if isinstance(k, int):
            return self.data[k]
        D = np.zeros(self.size)
        for x in k:
            D += self.data[x]
        return D

    def listChannels(self):
        return self.Peaks

    def show(self, ax, *channels):
        D = self.getChannels(*channels)
        ax.imshow(D)
        if len(channels) > 1:
            ax.set_title(",".join([str(z) for z in channels]))
        else:
            ax.set_title(str(channels[0]))



# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

# electron mass
me = 0.00054858  # u

# Electron charge
qe = 1.60217733e-19  # A*s

# Avogadro Number
NA = 6.0221367e23  # 1/mol

# Boltzmann
kb = 1.38064852e-23  # J/K

# gas constant
R = 8.3144598  # J/K

# factor between the sigma and FWHM for Gaussian
gfact = 2.3548200450309493



# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>


from . import utils
from ._version import __version__  # noqa: F401
from .Block import Block
from .Bruker import Bruker
from .collection import Collection
from .ITA import ITA, ITA_collection
from .ITAX import ITAX
from .ITM import ITM
from .ITS import ITS
from .nanoscan import Nanoscan
from .PCA import PCA
from .SPM import SPM_image
from .SXM import SXM

__all__ = [
    "ITA",
    "ITAX",
    "ITS",
    "ITM",
    "PCA",
    "Block",
    "SPM_image",
    "Collection",
    "Bruker",
    "Nanoscan",
    "utils",
    "SXM",
    "ITA_collection",
]



import numpy as np

from .math import clip01, positify, strictly_positify


def psf(img, sx, sy=None, angle=0):
    """
    Return a Gaussian PSF of the same size as img.

    img: image (reference for the output size)
    sx: sigma value for the long axis
    sy: sigma value for the short axis. If None take the same value as sx [default]
    angle: geometric angle (in radian) of the long axis. [default: 0]
    """
    from .math import Gauss

    if sy is None:
        sy = sx
    x = np.arange(img.shape[1])
    y = np.arange(img.shape[0])
    X, Y = np.meshgrid(x, y)
    X -= img.shape[1] // 2
    Y -= img.shape[0] // 2
    if angle != 0:
        Xp = X * np.cos(angle) - Y * np.sin(angle)
        Yp = X * np.sin(angle) + Y * np.cos(angle)
    else:
        Xp = X
        Yp = Y
    return Gauss(Xp, 0, sx) * Gauss(Yp, 0, sy)


def _rl(x, image, psf, type="default", extend=True, damping=0, ndamp=10):
    """
    Richardson-Lucy core algorithm
    Reference: L. B. Lucy / The Astronomical Journal / vol. 79 / No. 6 / June 1974 / pp. 745-754

    By giving an estimate x_k this function returns the next estimate x_{k+1}.

    x: x_k estimate
    image: input image to enhance
    psf: point spread functional
    """
    I = strictly_positify(
        convolve(x, psf, type=type, extend=extend)
    )  # reconvoluted estimation.
    if damping != 0:
        ratio = _rl_damped(I, image, damping=damping, ndamp=ndamp)
    else:
        ratio = image / I
    return x * convolve(
        ratio, psf[::-1, ::-1], type=type, extend=extend
    )  # Correlation is the convolution of mirrored psf


def _rl_damped(I, image, gain=1, con_var=1, damping=1, ndamp=10):
    """Calculate the damping ratio

    Parameters
    ----------
    gain: float, int
        CCD gain (relic?)
    con_var: float, int, np.ndarray
        Noise value or image
    threshold: float, int
        noise sigma threshold for dampening
    ndamp: float, int
        order of the dampening
    """

    from .haar import hfilter

    rrr = image - I
    rrr = hfilter(rrr, (I + con_var) / gain, damping, ndamp=ndamp)
    rrr[np.isnan(rrr)] = 0
    ratio = gain * (1 + rrr / (I + con_var))
    return ratio


def _rl_accelerate(x, x1, x2, g1=None, g2=None, order=1):
    """
    Accelerated Richardson-Lucy algorithm.
    Reference: David S. C. Biggs and Mark Andrews, Appl. Opt./ Vol. 36 / No. 8 / 10 March 1997 / pp. 1766-1775

    Notation in reference to paper:
    x  = x_k
    x1 = x_{k-1}
    x2 = x_{k_2}
    g1 = g_{k-1}
    g2 = g_{k-2}
    y  = y_k
    """

    if g2 is None:
        alpha = 0  # Initialization
    else:
        alpha = np.sum(g1 * g2) / strictly_positify(np.sum(g2**2))  # Eq. 10
        alpha = clip01(alpha)  # be sure α∈[0,1]
    if alpha == 0:
        return x  # the prediction is the same as x (initialization)
    h1 = x - x1  # Eq. 7
    y = x + alpha * h1  # Eq. 6
    if order > 1:
        h2 = x - 2 * x1 + x2  # Eq. 17
        y += h2 * alpha**2 / 2  # Eq. 14
    return y


def richardson_lucy(
    image,
    psf,
    iterations,
    damping=0,
    ndamp=10,
    core="default",
    acceleration=2,
    init="mean",
    extend=True,
    clip=False,
    **kargs,
):
    """
    Richardson-Lucy algorithm
    image: the image to enhance (numpy 2d array)
    psf: the Point Spread Function (numpy 2d array)
    iterations:
        The number of iterations to perform. It can be either an integer or a list of integer.
        For the later case, the returned solution is a dictionary with keys K and value being the enhancement after K interations.
    T: Damping factor ( to be used with core='damped' )
    N: N factor used with core='damped'
    core:
        default: default R-L algorithm using convolve from scipy.signal
        fft: performs a fftconvolution
    acceleration:
        0: (No acceleration. standard R-L)
        1: First order acceleration
        2: Second order acceleration
        higher orders are not yet implemented
    damping:
        damping factor. (0= no damping)
    init:
        'mean': the default. The start value for x is the mean of the image
        'image': the start value x is the image itself
        numpy array: if init is a 2d numpy array, its value will be used as init value for x
    """
    assert core in ["default", "fft", "accurate"]

    image = image.astype(np.float)
    psf = psf.astype(np.float)
    psf /= np.sum(psf)  # Normalize the psf ⇒ ∫∫ psf(x,y) dx dy = 1

    if init == "mean":
        x = 0.5 * np.ones(image.shape)
    elif init == "image":
        x = image
    else:
        x = init

    # Is iterations a number of a list of number?
    dict_output = True
    if isinstance(iterations, int):
        dict_output = False
        iterations = [iterations]

    N = max(iterations)

    results = {}
    x1 = x2 = None
    g1 = g2 = None

    for i in range(N):
        y = _rl_accelerate(x, x1, x2, g1, g2, order=acceleration) if acceleration else x
        x_new = _rl(
            positify(y),
            image=image,
            psf=psf,
            extend=extend,
            type=core,
            damping=damping,
            ndamp=ndamp,
        )
        g2 = g1
        g1 = x_new - y
        x, x1, x2 = x_new, x, x1  # rotate elements for next iteration
        if clip:
            x[x < 0] = 0
            x[x > clip] = clip
        if i + 1 in iterations:
            results[i + 1] = np.copy(x)
    if dict_output:
        return results
    return results[N]


def img_extend(img, margin, block=1):
    I = np.pad(img, margin, "constant")
    for i in range(img.shape[1]):
        I[:margin, i + margin] = np.mean(img[:block, i])
        I[-margin:, i + margin] = np.mean(img[-block:, i])
    for i in range(img.shape[0]):
        I[i + margin, :margin] = np.mean(img[i, :block])
        I[i + margin, -margin:] = np.mean(img[i, -block:])
    I[:margin, :margin] = np.mean(img[:block, :block])
    I[:margin, -margin:] = np.mean(img[:block, -block:])
    I[-margin:, :margin] = np.mean(img[-block:, :block])
    I[-margin:, -margin:] = np.mean(img[-block:, -block:])
    return I


def convolve(
    img, psf, type="default", extend=True, mode="same", extend_margin=100, **kargs
):
    """
    Compute the convolution of two 2D signals: img and psf
    type:
        define the convolution type
    """
    if extend is int:
        extend_margin = extend

    if extend:
        img = img_extend(img, extend_margin)

    if type == "fft":
        from scipy.signal import fftconvolve as conv

        I = conv(img, psf, mode)
    elif type == "default":
        from scipy.signal import convolve as conv

        I = conv(img, psf, mode)
    elif type == "accurate":
        from scipy.signal import convolve2d as convolve

        I = convolve(img, psf, mode)
    elif type == "fft2":
        I = np.fft.fftshift(np.fft.irfft2(np.fft.rfft2(img) * np.fft.rfft2(psf)))

    if extend:
        I = I[extend_margin:-extend_margin, extend_margin:-extend_margin]

    return I



# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

"""
collection is a module to handle collection of images.
This is specially useful for SPM data which store several channels for the same measurement
"""

import copy
import re

import matplotlib.pyplot as plt
import numpy as np

from . import SPM
from .SPM import SPM_image


def atoi(text):
    """
    Convert string to integer
    """
    return int(text) if text.isdigit() else text


def natural_keys(text):
    """
    alist.sort(key=natural_keys) sorts in human order
    http://nedbatchelder.com/blog/200712/human_sorting.html
    (See Toothy's implementation in the comments)
    """
    return [atoi(c) for c in re.split("(\\d+)", text)]


class Collection:
    """
    Class to handle a collection of SPM images
    """

    def __init__(self, sx=None, sy=None, unit="px", name="RawData", cls=None):
        """
        Create a new collection.
        You should provide a size.
        sx for the x-axis
        sy for the y-axis
        and unit for the name of the dimensional units
        """

        if isinstance(cls, Collection):
            self.size = cls.size
            self.name = cls.name

        if sy is None and sx is not None:
            sy = sx

        self.name = name
        self.channels = {}
        self.size = {"x": sx, "y": sy, "unit": unit}

    def add(self, Img, name, force=False):
        """
        Add a new Img to the collection
        """

        if len(self.channels) == 0 and self.size["unit"] == "px":
            if isinstance(Img, SPM_image):
                self.size = {"x": Img.pixels.shape[1], "y": Img.pixels.shape[0]}
            else:
                self.size["x"] = len(Img[0])
                self.size["y"] = len(Img)
        if name in self.channels:
            if force:
                del self.channels[name]
            else:
                raise KeyError(
                    "The channel {} is already present in "
                    "the collection. Please delete it before"
                )
        self.channels[name] = Img

    def create_image(self, img, key=None):
        return SPM_image(img, _type=self.name, real=self.size, channel=key)

    def __len__(self):
        return len(self.channels)

    def __getitem__(self, key):
        if key not in self.channels:
            return None
        if isinstance(self.channels[key], SPM_image):
            return self.channels[key]
        return self.create_image(self.channels[key], key=key)

    def __delitem__(self, key):
        del self.channels[key]

    def __iter__(self):
        self.iterator = iter(self.channels)
        return self

    def __next__(self):
        it = self.iterator.__next__()
        return self.__getitem__(it)

    def __setitem__(self, key, value):
        self.add(value, key, force=True)

    def show(self, ax=None, channels=None, cmap="hot", ncols=3, width=20, **kargs):
        """
        Display the images of the collection in a matplotlib plot.

        ax: is the axis of the matplotib plot.
            If None is provided, the current axis will be retrieved (or new one)
        channels: The channels to plot.
            If None, all will be plotted
        cmap: The colormap to use (hot)
        **kargs: further arguments will be passed to the show function of the generated SPM_Image
        """
        if channels is None:
            channels = list(self.channels.keys())
        channels_number = len(channels)
        assert channels_number > 0
        channels.sort(key=natural_keys)
        if ax is None:
            if channels_number == 4 and ncols == 3:
                fig, ax = plt.subplots(
                    2,
                    2,
                    figsize=(
                        width,
                        self[channels[0]].pixels.shape[0]
                        * width
                        / self[channels[0]].pixels.shape[1],
                    ),
                )
            else:
                Ny = (channels_number - 1) // ncols + 1
                Nx = min(ncols, channels_number)
                fig, ax = plt.subplots(
                    Ny,
                    Nx,
                    figsize=(width, ((channels_number - 1) // ncols + 1) * width / Nx),
                )
        if isinstance(ax, list):
            ax = np.array(ax).ravel()
        for i, x in enumerate(channels):
            self[x].show(ax=ax[i], cmap=cmap, **kargs)
        plt.tight_layout()

    def get_multivariate(self, channels=None):
        """
        Create and return a (pandas) DataFrame of the collection
        channels: List of the channels to use (default: None => all channels)

        Note: The images will be unraveled (flattened)
        """
        import pandas as pd

        if channels is None:
            channels = self.channels.keys()
        if isinstance(next(iter(self.channels.values())), SPM_image):
            data = pd.DataFrame({k: self.channels[k].pixels.ravel() for k in channels})
        else:
            data = pd.DataFrame({k: self.channels[k].ravel() for k in channels})
        return data

    def overlay(
        self,
        channel_names,
        colors=None,
        sig=None,
        vmin=None,
        vmax=None,
        **kargs,
    ):
        """
        Create an overlay (in color) of several channels.
        channel_names: a list of the channels to select for the overlay
        colors: List of [Red,Green,Blue] color to use.
            (Its length should be identical to channel_names)
        """
        if colors is None:
            colors = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
        assert len(channel_names) >= 2
        assert len(colors) >= len(channel_names)
        data = [
            SPM.normalize(self[ch].pixels, sig=sig, vmin=vmin, vmax=vmax)
            for ch in channel_names
        ]
        layers = [
            np.stack([data[i] * colors[i][j] for j in range(3)], axis=2)
            for i in range(len(channel_names))
        ]
        overlay = np.clip(np.sum(layers, axis=0), 0, 1)
        o = self.create_image(overlay, key="overlay")
        ch = [
            self.create_image(x, key=self[channel_names[i]].channel)
            for i, x in enumerate(layers)
        ]
        if "ax" in kargs:
            o.show(**kargs)
        return o, ch

    def stitch_correction(self, channel, stitches):
        """
        Function to correct for anomalies seen in stitched image.
        The function will calculate an average distribution of the stitched field
        and average the image with it and return a new collection with the result

        channel: name of the channel used as a reference
            (take one with homogeneous intensities in the whole image)
        stitches: a tuple/list containing the number of stitches in the image (x,y)
        """
        result = copy.deepcopy(self)
        del result.channels
        result.channels = {}
        size = next(iter(self.channels.values()))
        S = np.zeros((size[0] / stitches[0], size[1] / stitches[1]))
        for i in range(stitches[0]):
            for j in range(stitches[1]):
                # TODO: sx and sy are undefined
                S += self.channels[channel][
                    sy * i : sy * (i + 1), sx * j : sx * (j + 1)
                ]
        S[S == 0] = 1
        sy, sx = S.shape
        for x in self.channels:
            F = np.zeros(size)
            for i in range(stitches[0]):
                for j in range(stitches[1]):
                    F[sy * i : sy * (i + 1), sx * j : sx * (j + 1)] = (
                        self.channels[x][sy * i : sy * (i + 1), sx * j : sx * (j + 1)]
                        / S
                    )
            result.add(F, x)
        return result


def __Tsign(p1, p2, p3):
    return (p1[0] - p3[0]) * (p2[1] - p3[1]) - (p2[0] - p3[0]) * (p1[1] - p3[1])


def PointInTriangle(pt, v1, v2, v3):
    """
    Is a point pt in the triangle formed by the vertices v1, v2 and v3?
    pt,v1,v2,v3: tuple/list contai9ng the (x,y) coordinate of each vertices/point
    """
    b1 = __Tsign(pt, v1, v2) < 0
    b2 = __Tsign(pt, v2, v3) < 0
    b3 = __Tsign(pt, v3, v1) < 0
    return (b1 == b2) * (b2 == b3)


def overlay_triangle(
    channel_names,
    colors=None,
    radius=0.8,
    proportion=0.8,
    ax=None,
    size=512,
    bgcolor=None,
    textcolor="white",
    fontsize=20,
):
    """
    Create the image of a triangle, where the color is a gradient between three colors
    (one for each vertex).

    channel_names: a list of 3 channels
    colors: a list of 3 [R,G,B] list color
    ax: the matplotlib axis to plot in (if none use plt.gca())
    size: size of the rastered image generated.
        As there are in RGB no more than 256 values,
        the default value of 512 should be good in most of the cases.
    """
    if bgcolor is None:
        bgcolor = [0, 0, 0]
    if colors is None:
        colors = [[1, 0, 0], [0, 1, 0], [0, 0, 1]]
    assert len(channel_names) == 3
    assert len(colors) == 3
    if ax is None:
        ax = plt.gca()
    RGB = [bgcolor[i] * np.ones((size, size)) for i in range(3)]
    distance = 2 * radius * proportion * np.sin(np.radians(120))

    x = np.linspace(-0.7, 1.1, size)
    y = np.linspace(-1, 1, size)
    X, Y = np.meshgrid(x, y)
    centers = [
        (
            radius * proportion * np.cos(np.radians(120 * i)),
            radius * proportion * np.sin(np.radians(120 * i)),
        )
        for i in range(3)
    ]
    dist = [
        np.sqrt((X - centers[i][0]) ** 2 + (Y - centers[i][1]) ** 2) for i in range(3)
    ]

    for j in range(3):
        RGB[j] = sum(
            [
                colors[i][j] * np.maximum((distance - dist[i]) / distance, 0)
                for i in range(3)
            ]
        )
        ax.annotate(
            channel_names[j],
            (
                radius * np.cos(np.radians(120 * j)),
                radius * np.sin(np.radians(120 * j)),
            ),
            color=textcolor,
            fontsize=fontsize,
            va="center",
            ha="center",
        )
        RGB[j][PointInTriangle([X, Y], *centers) == 0] = bgcolor[j]
    image = np.stack(RGB, axis=2)
    ax.imshow(image, extent=[x[0], x[-1], y[-1], y[0]])
    ax.set_xticks([])
    ax.set_yticks([])



class Point:
    def __init__(self, xy, y=None):
        if y is None:
            assert type(xy) in [list, tuple]
            assert len(xy) == 2
            self.x = xy[0]
            self.y = xy[1]
        else:
            self.x = xy
            self.y = y

    def __add__(self, other):
        assert isinstance(other, Point)
        return Point(self.x + other.x, self.y + other.y)

    def __sub__(self, other):
        assert isinstance(other, Point)
        return Point(self.x - other.x, self.y - other.y)

    def __mul__(self, other):
        assert isinstance(other, Point)
        return Point(self.x * other.x, self.y * other.y)

    def __div__(self, other):
        assert isinstance(other, Point)
        return Point(self.x / other.x, self.y / other.y)


class Bbox:
    def __init__(self, *args, **kargs):
        if len(args) == 0:
            self.left = kargs["left"]
            self.right = kargs["right"]
            self.top = kargs["top"]
            self.bottom = kargs["bottom"]
        elif len(args) == 1:
            assert isinstance(args[0], dict)
            self.left = args[0]["left"]
            self.right = args[0]["right"]
            self.top = args[0]["top"]
            self.bottom = args[0]["bottom"]
        elif len(args) == 3:
            assert isinstance(args[0], (list, tuple))
            self.left = args[0][0]
            self.bottom = args[0][1]
            self.right = self.left + args[1]
            self.top = self.bottom + args[2]
        else:
            assert len(args) == 4
            self.left = args[0]
            self.right = args[1]
            self.top = args[2]
            self.bottom = args[3]

    def __repr__(self):
        return "Bbox ({left},{bottom}) -> ({right},{top})".format(**self.__dict__)

    def is_overlapping(self, other):
        return (
            (other.left < self.right)
            and (other.bottom < self.top)
            and (other.right > self.left)
            and (other.top > self.bottom)
        )

    def overlap(self, other):
        if not self.is_overlapping(other):
            return Bbox(0, 0, 0, 0)
        left = max(self.left, other.left)
        right = min(self.right, other.right)
        top = min(self.top, other.top)
        bottom = max(self.bottom, other.bottom)
        return Bbox(left, right, top, bottom)

    def show(self, ax=None, **kargs):
        import matplotlib as mpl
        import matplotlib.pyplot as plt

        from .plot import pixel2img

        if ax is None:
            ax = plt.gca()

        pos = pixel2img((self.left, self.bottom), ax=ax)
        ur = pixel2img((self.right, self.top), ax=ax)
        width = ur[0] - pos[0]
        height = ur[1] - pos[1]
        fill = kargs.pop("fill", None)
        color = kargs.pop("color", "r")
        ax.add_patch(
            mpl.patches.Rectangle(pos, width, height, fill=fill, color=color, **kargs)
        )



# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

"""
This is a standalone script which allows the user to perfome cross-section on ToF-SIMS images on different channels
"""

import os
import sys

import numpy as np
from PyQt5.QtCore import QCoreApplication
from PyQt5.QtWidgets import (
    QAction,
    QApplication,
    QFileDialog,
    QTableWidgetItem,
    QWidget,
)
from scipy import optimize as opt
from slicer import Ui_slicer

import pySPM
from pySPM.utils import CDF


class SlicerApp(QWidget):
    def __init__(self, filename=None):
        super(QWidget, self).__init__()
        self.ui = Ui_slicer()
        self.ui.setupUi(self)

        self.canvas = self.ui.mpl.canvas
        self.fig = self.ui.mpl.canvas.fig
        self.initPlotLayout()
        self.level = None
        if filename is None:
            if len(sys.argv) < 2:
                self.path, _ = QFileDialog.getOpenFileName(
                    self, "ITA Image", "", "(*.ITA)"
                )
            else:
                # If an argument is sent to the script, the first argument will be used as a Path. Very usefull for debugging the script without having to selectr the folder each time with window dialog
                self.path = sys.argv[1]
        else:
            self.path = filename
        if not os.path.exists(self.path):
            raise Exception(f'File "{self.path}" is not found')
        if os.path.exists(self.path + ".level.npy"):
            self.level = np.load(self.path + ".level.npy")
        self.curs = [0, 0, 0]
        self.volume = None
        self.ITA = pySPM.ITA(self.path)
        for i, x in enumerate(self.ITA.get_masses()):
            self.ui.peakList.setRowCount(i + 1)
            self.ui.peakList.setItem(i, 0, QTableWidgetItem(x["assign"]))
            self.ui.peakList.setItem(
                i, 1, QTableWidgetItem("{:.2f}u".format(x["cmass"]))
            )
            self.ui.peakList.setItem(
                i, 2, QTableWidgetItem("{:.2f}u".format(x["umass"] - x["lmass"]))
            )
        self.ui.peakList.show()
        self.ui.cmap.currentIndexChanged.connect(self.plot)
        self.ui.prof1daxis.currentIndexChanged.connect(self.plot)
        self.ui.peakList.cellClicked.connect(self.load_channel)
        self.canvas.mpl_connect("button_press_event", self.on_pick)
        self.flatAction = QAction("Flatten substrate from this channel")
        self.flatAction.triggered.connect(self.flatten)
        self.ui.correction.stateChanged.connect(self.plot)
        self.ui.peakList.addAction(self.flatAction)
        self.ui.status.setText("IDLE")

    def flatline(self, y):
        for x in range(self.volume.shape[1]):
            popt, pcov = opt.curve_fit(
                CDF,
                np.arange(self.volume.shape[2]),
                self.volume[y, x, :],
                (10, self.volume.shape[2] / 2, 1),
            )
            self.level[y, x] = popt[1]

    def flatten(self):
        self.ui.status.setText("Start the flattening...")
        self.level = np.zeros(self.volume.shape[:2])
        self.ui.pb.setMaximum(self.volume.shape[0])
        for y in range(self.volume.shape[0]):
            self.ui.pb.setValue(y)
            self.flatline(y)
            QCoreApplication.processEvents()
            self.ax.clear()
            self.ax.imshow(self.level)
            self.canvas.draw()
        self.ui.pb.setValue(0)
        self.ui.status.setText("Flattening finished")
        np.save(self.path + ".level", self.level)

    def load_channel(self, row, col):
        self.ui.status.setText("Loading channel...")
        id = row
        vol = []
        for i in range(self.ITA.Nscan):
            x = self.ITA.getImage(id, i)
            vol.append(x)
        self.volume = np.stack(list(vol), axis=2)
        if self.level is not None:
            self.corrected = np.zeros(self.volume.shape)
            np.arange(self.ITA.Nscan)
            self.ui.pb.setMaximum(self.level.shape[0])
            for y in np.arange(self.level.shape[0]):
                self.ui.pb.setValue(y)
                for x in np.arange(self.level.shape[1]):
                    dz = int(-self.level[y, x] + np.max(self.level))
                    self.corrected[y, x, dz:] = self.volume[
                        y, x, : self.volume.shape[2] - dz
                    ]
            self.ui.pb.setValue(0)
        self.plot()
        self.ui.status.setText("IDLE")

    def plot(self):
        A = self.corrected if self.ui.correction.isChecked() else self.volume
        cmap = self.ui.cmap.currentText()
        self.axXY.clear()
        self.axXZ.clear()
        self.axYZ.clear()
        self.ax.clear()
        if self.volume is None:
            return
        self.axXY.imshow(A[:, :, self.curs[2]], cmap=cmap)
        self.axXZ.imshow(A[self.curs[1], :, :].T, cmap=cmap)
        self.axYZ.imshow(A[:, self.curs[0], :].T, cmap=cmap)
        self.axXY.axhline(self.curs[1])
        self.axXY.axvline(self.curs[0])
        self.axXZ.axhline(self.curs[2])
        self.axXZ.axvline(self.curs[0])
        self.axYZ.axhline(self.curs[2])
        self.axYZ.axvline(self.curs[1])
        self.axXY.set_title("XY")
        self.axXZ.set_title("XZ")
        self.axYZ.set_title("YZ")
        if self.ui.prof1daxis.currentText() in "XYZ":
            i = "XYZ".index(self.ui.prof1daxis.currentText())
            self.ax.set_xlabel("XYZ"[i])
            if i == 0:
                self.ax.plot(A[:, self.curs[0], self.curs[2]])
            elif i == 1:
                self.ax.plot(A[self.curs[1], :, self.curs[2]])
            elif i == 2:
                self.ax.plot(A[self.curs[1], self.curs[0], :])
        self.canvas.draw()

    def on_pick(self, event):
        if event.inaxes not in [self.axYZ, self.axXZ, self.axXY]:
            return
        x = event.xdata
        y = event.ydata
        [self.axYZ, self.axXZ, self.axXY].index(event.inaxes)
        xdata = int(x)
        ydata = int(y)

        if event.inaxes == self.axXY:
            self.curs[0] = xdata
            self.curs[1] = ydata
        elif event.inaxes == self.axYZ:
            self.curs[1] = xdata
            self.curs[2] = ydata
        elif event.inaxes == self.axXZ:
            self.curs[0] = xdata
            self.curs[2] = ydata
        else:
            print("Click event not handled")
        self.curs = [
            np.clip(0, self.curs[i], self.volume.shape[i] - 1) for i in range(3)
        ]
        self.plot()

    def initPlotLayout(self):
        """
        Setup the plotting layout.
        """
        self.axXY = self.fig.add_subplot(2, 2, 1)
        self.axYZ = self.fig.add_subplot(2, 2, 2)
        self.axXZ = self.fig.add_subplot(2, 2, 3)
        self.ax = self.fig.add_subplot(2, 2, 4)
        self.fig.tight_layout()


app = QApplication(sys.argv)
window = SlicerApp()
window.show()
sys.exit(app.exec_())



# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

import os.path

from .Block import Block


class InvalidRawDataFormatError(Exception):
    def __init__(self, block, msg):
        self.block = block
        self.msg = msg

    def __str__(self):
        return (
            "Invalid RAW dataformat seen in block "
            + self.block.parent
            + "/"
            + self.block.name
            + " : "
            + self.msg
        )


class ITS:
    def __init__(self, filename, debug=False):
        """
        ITS
        """
        self.filename = filename
        assert os.path.exists(filename)
        self.f = open(self.filename, "rb")
        self.Type = self.f.read(8)
        assert self.Type == b"ITStrF01"
        self.root = Block(self.f)



# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

"""
Module to handle SPM images recorded by a Bruker AFM
"""

import contextlib
import re
import struct

import numpy as np

import pySPM


class Bruker:
    """
    Class to handle SPM images recorded by a Bruker AFM
    """

    def __init__(self, path):
        self.path = path
        with open(self.path, "rb") as file:
            self.layers = []
            self.scanners = []
            mode = ""
            while True:
                line = file.readline().rstrip().replace(b"\\", b"")
                if line == b"*Ciao image list":
                    self.layers.append({})
                    mode = "Image"
                elif line == b"*Scanner list":
                    self.scanners.append({})
                    mode = "Scanner"
                elif line.startswith(b"*EC"):
                    mode = "EC"
                else:
                    args = line.split(b": ")
                    if len(args) > 1:
                        if mode == "Image":
                            self.layers[-1][args[0]] = args[1:]
                        elif mode == "Scanner":
                            self.scanners[-1][args[0]] = args[1:]
                    if line == b"*File list end":
                        break

    def _get_bpp(self, i):
        int(self.layers[i][b"Data offset"][0])
        cols = int(self.layers[i][b"Number of lines"][0])
        rows = int(self.layers[i][b"Samps/line"][0])
        byte_length = int(self.layers[i][b"Data length"][0])
        length = rows * cols
        bpp = byte_length // length
        return bpp

    def _get_raw_layer(self, i, debug=False, mock_data=False):
        """
        Internal function to retrieve raw data of a layer
        """
        off = int(self._get_layer_val(i, "Data offset"))
        if debug:
            print("RAW offset: ", off)
        cols, rows = self._get_res(i)
        byte_length = int(self._get_layer_val(i, "Data length"))
        length = rows * cols
        bpp = byte_length // length
        byte_length = length * bpp
        if mock_data:
            return np.zeros((rows, cols))

        with open(self.path, "rb") as file:
            file.seek(off)
            return np.array(
                struct.unpack(
                    "<" + str(length) + {2: "h", 4: "i", 8: "q"}[bpp],
                    file.read(byte_length),
                ),
                dtype="float64",
            ).reshape((rows, cols))

    def list_channels(self, encoding="latin1"):
        print("Channels")
        print("========")

        for layer in self.layers:
            with contextlib.suppress(KeyError):
                print(layer[b"@2:Image Data"][0].decode(encoding))
        for layer in self.layers:
            with contextlib.suppress(KeyError):
                print(layer[b"@3:Image Data"][0].decode(encoding) + " (MFM)")

    def _get_layer_val(self, index: int, name: str, first=True):
        lname = name.lower()
        lname2 = name[0] + lname[1:]
        if name.encode() in self.layers[index]:
            val = self.layers[index][name.encode()]
        elif lname.encode() in self.layers[index]:
            val = self.layers[index][lname.encode()]
        elif lname2.encode() in self.layers[index]:
            val = self.layers[index][lname2.encode()]
        if first:
            return val[0]
        return val

    def _get_res(self, layer_index):
        row_key = (
            "Valid data len X"
            if b"Valid data len X" in self.layers[layer_index]
            else "Number of lines"
        )
        col_key = (
            "Valid data len Y"
            if b"Valid data len Y" in self.layers[layer_index]
            else "Samps/line"
        )
        xres = int(self._get_layer_val(layer_index, row_key))
        yres = int(self._get_layer_val(layer_index, col_key))
        return xres, yres

    def _get_layer_size(self, layer_index, encoding, debug=False):
        scan_size = self._get_layer_val(layer_index, "Scan Size").split()
        xres, yres = self._get_res(layer_index)
        if scan_size[2][0] == 126:
            scan_size[2] = b"u" + scan_size[2][1:]
        size = {}
        size = {
            "x": float(scan_size[0]),
            "y": float(scan_size[1]) * yres / xres,
            "unit": scan_size[2].decode(encoding),
        }
        if debug:
            print("scan size", scan_size)

        return size

    def get_channel(
        self,
        channel="Height Sensor",
        backward=False,
        corr=None,
        debug=False,
        encoding="latin1",
        lazy=True,
        mfm=False,
        mock_data=False,
    ):
        """
        Load the SPM image contained in a channel
        """
        for i in range(len(self.layers)):
            if mfm:
                # Handle case of MFM, where image data is stored at layer 3 on the backward channel
                backward = True
                _type = "Bruker MFM"
                try:
                    layer_name = self._get_layer_val(i, "@3:Image Data").decode(
                        encoding
                    )
                except KeyError:
                    continue
            else:
                _type = "Bruker AFM"
                try:
                    layer_name = self._get_layer_val(i, "@2:Image Data").decode(
                        encoding
                    )
                except KeyError:
                    continue
            result = re.match(r'([^ ]+) \[([^]]*)] "([^"]*)"', layer_name).groups()
            if result[2] == channel:
                if debug:
                    print("channel " + channel + " Found!")
                bck = False
                if self._get_layer_val(i, "Line Direction") == b"Retrace":
                    bck = True
                if bck == backward:
                    if debug:
                        print("Direction found")
                    var = self._get_layer_val(i, "@2:Z scale").decode(encoding)
                    if debug:
                        print("@2:Z scale", var)
                    if "[" in var:
                        result = re.match(
                            r"[A-Z]+\s+\[([^]]+)]\s+\(-?[0-9.]+ .*?\)\s+(-?[0-9.]+)\s+(.*?)$",
                            var,
                        ).groups()
                        if debug:
                            print(result)
                        bpp = int(self._get_layer_val(i, "Bytes/pixel"))
                        if debug:
                            print("BPP", bpp)
                        # scale = float(result[1])
                        scale = float(result[1]) / 256**bpp

                        result2 = self.scanners[0][b"@" + result[0].encode(encoding)][
                            0
                        ].split()
                        if debug:
                            print("result2", result2)
                        scale2 = float(result2[1])
                        zscale = result2[2] if len(result2) > 2 else result2[0]
                        if b"/V" in zscale:
                            zscale = zscale.replace(b"/V", b"")
                        if debug:
                            print(f"scale: {scale:.3e}")
                            print(f"scale2: {scale2:.3e}")
                            print("zscale: " + str(zscale))
                        var = self._get_layer_val(i, "@2:Z offset").decode(encoding)
                        result = re.match(
                            r"[A-Z]+\s+\[[^]]+]\s+\(-?[0-9.]+ .*?\)\s+(-?[0-9.]+)\s+.*?$",
                            var,
                        ).groups()
                        offset = float(result[0])
                    else:
                        if debug:
                            print("mode 2")
                        result = re.match(
                            r"[A-Z]+ \(-?[0-9.]+ [^)]+\)\s+(-?[0-9.]+) [\w]+", var
                        ).groups()
                        scale = float(result[0]) / 65536.0
                        scale2 = 1
                        zscale = b"V"
                        result = re.match(
                            r"[A-Z]+ \(-?[0-9.]+ .*?\)\s+(-?[0-9.]+) .*?",
                            self._get_layer_val(i, "@2:Z offset").decode(encoding),
                        ).groups()
                        offset = float(result[0])
                    if debug:
                        print("Offset:", offset)
                    data = (
                        self._get_raw_layer(i, debug=debug, mock_data=mock_data)
                        * scale
                        * scale2
                    )
                    xres, yres = self._get_res(i)
                    if debug:
                        print("xres/yres", xres, yres)
                    scan_size = self._get_layer_val(i, "Scan Size").split()
                    aspect_ratio = [
                        float(x)
                        for x in self._get_layer_val(i, "Aspect Ratio").split(b":")
                    ]
                    if debug:
                        print("aspect ratio", aspect_ratio)
                        print("scan size", scan_size)
                    if scan_size[2][0] == 126:
                        scan_size[2] = b"u" + scan_size[2][1:]
                    size = {
                        "x": float(scan_size[0]),
                        "y": float(scan_size[1]) * yres / xres,
                        "unit": scan_size[2].decode(encoding),
                    }
                    image = pySPM.SPM_image(
                        channel=[channel, "Topography"][channel == "Height Sensor"],
                        BIN=data,
                        real=size,
                        _type=_type,
                        zscale=zscale.decode(encoding),
                        corr=corr,
                    )
                    return image
        if lazy:
            return self.get_channel(
                channel=channel,
                backward=not backward,
                corr=corr,
                debug=debug,
                encoding=encoding,
                lazy=False,
                mock_data=mock_data,
            )
        raise Exception(f"Channel {channel} not found")



# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

"""
This module gives the ability to ready and parse the ToF-SIMS ITA files from iontof.
You can mainly retrieve images and spectra for each channel and scan.
"""

import os.path
import re
import struct
import zlib

import numpy as np

from .Block import MissingBlockError
from .collection import Collection
from .ITM import ITM
from .utils.misc import PB, alias, aliased, deprecated


@aliased
class ITA(ITM):
    def __init__(self, filename, *args, **kargs):
        """
        Open an ITA file.

        Parameters
        ----------
        filename : string
            the path of the ita file

        Returns
        -------
        Class<ITA>
            ITA Object

        Examples
        --------
        >>> import pySPM
        >>> filename = "myfile.ita"
        >>> A = pySPM.ITA(filename)
        """
        ITM.__init__(self, filename, *args, **kargs)
        try:
            self.sx = self.root.goto(
                "filterdata/TofCorrection/ImageStack/Reduced Data/ImageStackScans/Image.XSize"
            ).get_ulong()
            self.sy = self.root.goto(
                "filterdata/TofCorrection/ImageStack/Reduced Data/ImageStackScans/Image.YSize"
            ).get_ulong()
        except MissingBlockError:
            self.sx = self.size["pixels"]["x"]
            self.sy = self.size["pixels"]["y"]
        try:
            # self.Nscan = int(self.root.goto('filterdata/TofCorrection/ImageStack/Reduced Data'\
            #    '/ImageStackScans/Image.NumberOfScans').getLong())
            self.Nimg = int(
                self.root.goto(
                    "filterdata/TofCorrection/ImageStack/Reduced Data"
                    "/ImageStackScans/Image.NumberOfImages"
                ).get_ulong()
            )
        except:
            self.Nimg = 0
        self.img = self.get_intensity()

        try:
            self.fov = self.root.goto("Meta/SI Image[0]/fieldofview").get_double()
        except MissingBlockError:
            self.fov = self.get_value("Registration.Raster.FieldOfView")["float"]

    @alias("getIntensity")
    def get_intensity(self):
        """
        Retrieve the total Ion image
        """
        try:
            X, Y = self.size["pixels"]["x"], self.size["pixels"]["y"]
            img = self.image(
                np.flipud(
                    np.array(
                        self.root.goto("Meta/SI Image/intensdata").get_data("f"),
                        dtype=np.float32,
                    ).reshape((Y, X))
                ),
                channel="SI count",
            )
        except Exception:
            try:
                img = self.get_added_image(0).pixels
            except:
                try:
                    img = self.get_added_image_by_SN(self.get_channel_SN("total"))
                except:
                    import warnings

                    warnings.warn("SI image cannot be retrieved", stacklevel=2)
                    return None
        return img

    def get_channel_SN(self, channel):
        """
        New ITA fileformat assign a serial number (SN) in the form of a UUID for each channel.
        The SN corresponding to a given channel name can be retrieved by this function.

        Parameters
        ----------
        channel : string
            The channel name assigned to a given peak
        """
        for x in self.root.goto("MassIntervalList"):
            if x.name == "mi":
                l = x.dict_list()
                if l["assign"]["utf16"] == channel or l["desc"]["utf16"] == channel:
                    return l["SN"]["utf16"]

        raise Exception(f'Channel name "{channel}" not found')

    @alias("getChannelsByName")
    def get_channels_by_name(self, name, strict=False):
        """
        Retrieve the channels for a given assignment name in the form of a list of dictionaries.
        The output can be formatted in a human readable way with the pySPM.ITA.showChannels function (see examples).

        Parameters
        ----------
        name : string or list of strings
            A regular expression (regex) used for the search
        strict : bool
            If strict is True, the search name won't be treated as a regexp, but rather the whole name should match.

        Returns
        -------
        list
            A list of dictionaries where each dictionary is a description of the selected channel. Which contains:
                - clsid : class ID. A useless information for the end-user
                - desc : a description string encoded in utf16.
                - color : a 32bits color encoding of the peak
                - symbolID : Not used
                - id : The ID of the channel. (The total counts is 0, the sum
                  of the rest 1 and the first peak is 2, ... )
                - SN : an utf16 serial number which is useless for the end-used
                - assign : a utf16 string with the element assignment of the
                  peak (e.g.: CH-, Na+, ...)
                - lmass : a long value indicating the lower mass of the peak (in u)
                - umass : a long value indicating the upper mass of the peak (in u)
                - cmass : a long value indicating the center mass of the peak

        Examples
        --------
        >>> A = pySPM.ITA("myfile.ita")
        >>> ch = A.getChannelsByName("C")
        >>> A.showChannels(ch)
                CH- (), mass: 12.99 - 13.03
            C_2- (), mass: 23.97 - 24.03
            C_2H- (), mass: 24.98 - 25.04
            CN- (), mass: 25.97 - 26.04
            Cl- (), mass: 34.93 - 35.01
            C_2O- (), mass: 39.96 - 40.04
            CHNO- (), mass: 42.97 - 43.05
            CHO_2- (), mass: 44.95 - 45.04
            Cs- (), mass: 132.81 - 133.01
        >>> ch = A.getChannelsByName("C[^a-z]") # Only carbon atoms (meaning that the char after C cannot be lowercase)
        >>> A.showChannels(ch)
                CH- (), mass: 12.99 - 13.03
            C_2- (), mass: 23.97 - 24.03
            C_2H- (), mass: 24.98 - 25.04
            CN- (), mass: 25.97 - 26.04
            C_2O- (), mass: 39.96 - 40.04
            CHNO- (), mass: 42.97 - 43.05
            CHO_2- (), mass: 44.95 - 45.04
        >>> ch = A.getChannelsByName("CH", True) # Only CH channel and not CHNO and CHO_2
        >>> A.showChannels(ch)
                CH- (), mass: 12.99 - 13.03
        """
        res = []
        if strict:
            if type(name) in [list, tuple]:
                name = ["^" + n + "[+-]?$" for n in name]
            else:
                name = "^" + name + "[+-]?$"
        if not isinstance(name, list):
            name = [name]
        for n in name:
            for P in self.peaks:
                p = self.peaks[P]
                ma = re.compile(n, re.U)
                if ma.match(p["assign"]["utf16"]) or ma.match(p["desc"]["utf16"]):
                    res.append(p)
        return res

    @deprecated("showChannels")
    def show_channels(self, ch):
        """
        Format a list of channels where each channel is represented by a dictionary (like the ones produced by pySPM.ITA.getChannelsByName) to a human readable output.

        Parameters
        ----------
        ch : list
            A list of dictionaries representing the channels

        Returns
        -------
        None
            It will print a list of channels with the assignment, the description in parenthesis followed by the lower - upper mass range.
        """
        for z in ch:
            print(
                "\t{name} ({desc}), mass: {lower:.2f} - {upper:.2f}".format(
                    desc=z["desc"]["utf16"],
                    name=z["assign"]["utf16"],
                    lower=z["lmass"]["float"],
                    upper=z["umass"]["float"],
                )
            )

    @alias("getChannelByMass")
    def get_channel_by_mass(self, mass, full=False):
        """
        Retrieves the first channel ID which has a mass range containing a given mass.

        Parameters
        ---------
        mass : int, float
            The mass. If zero, the channel 0 will be returned and correspond to the Total count channel.
        full : bool
            If True, not only the ID is retrieved but the whole dictionary similarly as with pySPM.ITA.getChannelsByName

        Returns
        -------
        int
            The first channel ID containing the mass given in argument. If a mass 0 is given, the output will be 0 which corresponds to the total count channel.
        """
        if mass == 0:
            return 0
        for P in self.peaks:
            p = self.peaks[P]

            if (
                p["id"]["long"] > 1
                and p["lmass"]["float"] <= mass
                and mass <= p["umass"]["float"]
            ):
                if full:
                    return p
                return p["id"]["long"]
        raise ValueError(f"Mass {mass:.2f} Not Found")

    @alias("getShiftCorrectedImageByName")
    def get_shift_corrected_image_by_name(self, names, **kargs):
        """
        Retrieve the drift corrected (or shift corrected) image for the sum of all channels matching a given name. The shift correction applied is the one saved in the ITA file.

        Parameters
        ---------
        names : string or list of strings
            A channel name of a list of channel names to be summed up

        Returns
        -------
        pySPM.SPM.SPM_image
            The image of the sum of all the selected channels
        list of dictionaries
            The list of all the channels selected. This list can be displayed in a human readable form by the pySPM.ITA.showChannels function

        """
        return self.get_sum_image_by_name(
            names, shifts=[(-x, -y) for x, y in self.get_saved_shift()], **kargs
        )

    def __get_sum_image(self, scans, channels, **kargs):
        """
        An internal function to retrieve the sum of several scans for several channel ID.
        """
        Z = np.zeros((self.sy, self.sx))
        if "shifts" in kargs:
            shifts = kargs["shifts"]
        elif "Shifts" in kargs:
            shifts = kargs["Shifts"]
        else:
            shifts = [(-x, -y) for x, y in self.get_saved_shift()]
        for ch in channels:
            ID = ch["id"]["long"]
            Z += self.fast_get_image(ID, scans, shifts)
        return Z

    @alias("getSumImageBySN")
    def get_sum_image_by_sn(self, SN, scans=None, prog=False, raw=False, **kargs):
        """
        Retrieve the image for the sum of several scans for a given channel SN.
        """
        if scans is None:
            scans = range(self.Nscan)
        if isinstance(scans, int):
            scans = [scans]
        if prog:
            scans = PB(scans)

        Z = np.zeros((self.sy, self.sx))
        for s in scans:
            node = self.root.goto(
                f"filterdata/TofCorrection/ImageStack/Reduced Data/Images/{SN}/ScanData/EDROff/{s}"
            )
            dat = node.decompress()
            data = struct.unpack(f"<{len(dat) // 4}I", dat)
            Z += np.array(data, dtype=np.float64).reshape((self.sy, self.sx))
        if raw:
            return Z
        channel = self.get_channel_by_sn(SN)
        return self.image(np.flipud(Z), channel=channel)

    @alias("getSumImageByName")
    def get_sum_image_by_name(
        self, names, scans=None, strict=False, prog=False, raw=False, **kargs
    ):
        """
        Retrieve the image for the sum of several scans and channels selected by their channel name.

        Parameters
        ----------
        names : string or list of strings
            Similar as for pySPM.ITA.getChannelsByName
        scans : int, list of int or None
            The list of the scan number to be summed up. For the case of None (default) all the available scans are taken.
        strict : bool
            Is the name selection strict? (see pySPM.ITA.getChannelsByName)
        prog : bool
            If True a progressbar will be displayed to show the summing progress as this might be quite slow.
        raw : bool
            If True a numpy array will be returned instead of a pySPM.SPM.SPM_image
        """
        if scans is None:
            scans = range(self.Nscan)
        if isinstance(scans, int):
            scans = [scans]

        channels = self.get_added_image_by_name(names, strict)
        if prog:
            scans = PB(scans)
        Z = self.__get_sum_image(scans, channels)
        if raw:
            return Z, channels
        channel_title = ",".join([z["assign"]["utf16"] for z in channels])
        return self.image(np.flipud(Z), channel=channel_title), channels

    def show(self, ax=None):
        """
        Shows the total SI image with the indication of the field of view.

        Parameters
        ----------
        ax : matplotlib axis or None
            The axis in which the image will be shown. If None the current axis will be used (ax = plt.gca())

        Returns
        -------
        None
        """
        import matplotlib.pyplot as plt

        if ax is None:
            fig, ax = plt.subplots(1, 1, figsize=(5, 5))
        ax.imshow(self.img, extent=(0, self.fov * 1e6, 0, self.fov * 1e6))
        ax.set_title("Total SI")
        ax.set_xlabel(r"x [$\mu$m]")
        ax.set_ylabel(r"y [$\mu$m]")

    @alias("getShiftsByMass")
    def get_shifts_by_mass(self, masses, centered=True, prog=False, Filter=None):
        """
        Deprecated. A relic function that the developer is not even sure what it was supposed to do ;)
        """
        Shifts = [(0, 0)]
        if Filter is None:

            def Filter(z):
                return z

        S0 = Filter(self.get_added_image_by_mass(masses, 0))
        Y = range(1, self.Nscan)
        if prog:
            Y = PB(Y)
        for i in Y:
            S = Filter(self.get_sum_image_by_mass(masses, i))
            Shift = np.real(
                np.fft.fftshift(np.fft.ifft2(np.conj(np.fft.fft2(S0)) * np.fft.fft2(S)))
            )
            cord = np.unravel_index(np.argmax(Shift), S0.shape)
            trans = (cord[1] - S0.shape[1] / 2, cord[0] - S0.shape[0] / 2)
            Shifts.append(trans)
        if centered:
            avSx = np.round(np.mean([z[0] for z in Shifts]))
            avSy = np.round(np.mean([z[1] for z in Shifts]))
            Shifts = [(z[0] - avSx, z[1] - avSy) for z in Shifts]
        return Shifts

    @alias("getXsectionByMass")
    def get_xsection_by_mass(
        self,
        x1,
        y1,
        x2,
        y2,
        masses,
        N=None,
        prog=False,
        ax=None,
        flip=False,
        col="w-",
        **kargs,
    ):
        """
        Retrieves a Cross-Section for a given mass along the profile determined by coordinates (x1,y1) and (x2,y2).
        The output is a 2D image where the x-axis correspond to the position along the profile and the y-axis the scan number.

        Parameters
        ----------
        x1 : int
        y1 : int
        x2 : int
        y2 : int
            profile coordinates in pixel: (x1,y1) -> (x2,y2)
        masses : int, float, list of floats
            The masse or list of masses to sum
        N : int or None
            The number of value used along the profile (which will be interpolated).
            None (default) will take the roundest number of values closest to the pixel length of the profile
        prog : bool
            If True display a progressbar
        ax : None or matplotlib axis
            if not None, the axis representing the 2D image can be given in order to display the profile's position
        col : string (matplotlib color format)
            The color of the profile used in case ax is given
        flip : bool
            Flip the y-coordinates?
        **kargs : arguments
            All supplementary arguments are passed to the pySPM.ITA.getSumImageByMass

        Returns
        -------
        np.ndarray
            2D numpy array containing the sum of all channels. The values are the count number
        """
        y1 = self.sy - 1 - y1
        y2 = self.sy - 1 - y2
        if N is None:
            N = int(np.sqrt((x2 - x1) ** 2 + (y2 - y1) ** 2)) + 1
        x = np.linspace(x1, x2, N)
        y = np.linspace(y1, y2, N)
        out = np.zeros((self.Nscan, N))
        Y = range(self.Nscan)
        if ax is not None and not flip:
            ax.plot([x1, x2], [self.sy - 1 - y1, self.sy - 1 - y2], col)
        if prog:
            Y = PB(Y)
        from scipy.ndimage import map_coordinates

        for s in Y:
            Z = self.get_sum_image_by_mass(masses, s, **kargs)
            P = map_coordinates(Z.pixels, np.vstack((y, x)))
            out[s, :] = P
        return out

    @alias("getAddedImageByName")
    def get_added_image_by_name(self, names, strict=False, raw=False, **kargs):
        """
        Retrieve the image for the sum of all scan (precomputed by iontof, but not shift-corrected) for given names

        Parameters
        ----------
        names : string or list of strings
            name of the channel (see pySPM.ITA.getChannelsByName)
        strict : bool
            If True the names are the exact names (see pySPM.ITA.getChannelsByName)
        raw : bool
            If True a 2D numpy array will be returned
        **kargs: supplementary arguments
            passed to pySPM.ITA.getAddedImage

        Returns
        -------
        pySPM.SPM.SPM_image
            Image of the result
        list of dictionaries
            List of all selected peaks used to compute the image.
            Note: Pass this list to pySPM.ITA.showChannels in order to print a human readable representation of it.
        """
        Z = np.zeros((self.sy, self.sx))
        channels = self.get_channels_by_name(names, strict)
        for ch in channels:
            ID = ch["id"]["long"]
            Z += self.get_added_image(ID, **kargs)
        ch = self.get_masses(channels)
        if raw:
            return Z, ch
        return self.image(np.flipud(Z), channel=",".join([z["assign"] for z in ch])), ch

    @alias("getSavedShift")
    def get_saved_shift(self):
        """
        getSavedShift returns the shifts saved with the file. Usually this is the shift correction you perform with the IonToF software.

        Returns
        -------
        List of tuples
            each tuple is a (Δx,Δy) in pixels (one for each scan).
        """
        try:
            X = zlib.decompress(
                self.root.goto(
                    "filterdata/TofCorrection/ImageStack/Reduced Data"
                    "/ImageStackScans/ShiftCoordinates/ImageStack.ShiftCoordinates"
                ).value
            )
        except:
            return [(0, 0) for x in range(self.Nscan)]
        D = struct.unpack("<" + str(len(X) // 4) + "i", X)
        dx = D[::2]
        dy = D[1::2]
        return list(zip(dx, dy))

    @alias("getShiftCorrectedImageByMass")
    def get_shift_corrected_image_by_mass(self, masses, **kargs):
        """
        Shortcut function for pySPM.ITA.get_sum_image_by_mass using the saved shift corrections.
        """
        return self.get_sum_image_by_mass(
            masses, shifts=[(-x, -y) for x, y in self.get_saved_shift()], **kargs
        )

    @alias("getSumImageByMass")
    def get_sum_image_by_mass(self, masses, scans=None, prog=False, raw=False, **kargs):
        """
        Similar to pySPM.ITA.getSumImageByName but instead of the names, the mass or list of mass is provided
        see pySPM.ITA.getSumImageByName for more details
        """
        if scans is None:
            scans = range(self.Nscan)
        if isinstance(scans, int):
            scans = [scans]
        if isinstance(masses, (int, float)):
            masses = [masses]
        if prog:
            scans = PB(scans, leave=False)
        channels = [self.get_channel_by_mass(m, full=True) for m in masses]
        Z = self.__get_sum_image(scans, channels, **kargs)
        if raw:
            return Z, channels
        channels_name = [
            ["{:.2f}u".format(m["cmass"]["float"]), m["assign"]["utf16"]][
                m["assign"]["utf16"] != ""
            ]
            for m in channels
        ]
        return self.image(np.flipud(Z), channel="Masses: " + ",".join(channels_name))

    @alias("getAddedImageByMass")
    def get_added_image_by_mass(self, masses, raw=False, **kargs):
        """
        Retrieve the image for the sum of all scan (precomputed by iontof, but not shift-corrected) for (a) given masse(s)

        Parameters
        ----------
        masses : float or list of float
            mass of the channels to be used
        raw : bool
            If True a 2D numpy array will be returned
        **kargs: supplementary arguments
            passed to pySPM.ITA.getAddedImage

        Returns
        -------
        pySPM.SPM.SPM_image
            Image of the result
        list of dictionaries
            Only returned if raw is True
            List of all selected peaks used to compute the image.
            Note: Pass this list to pySPM.ITA.showChannels in order to print a human readable representation of it.
        """
        if isinstance(masses, (int, float)):
            masses = [masses]
        Z = np.zeros((self.sy, self.sx))
        channels = []
        for m in masses:
            ch = self.get_channel_by_mass(m)
            m = self.get_masses()[ch]
            if m["assign"] != "":
                channels.append(m["assign"])
            else:
                channels.append("{cmass:.2f}u".format(**m))
            Z += self.get_added_image(ch, **kargs)
        if raw:
            return Z, channels
        return self.image(np.flipud(Z), channel=",".join(channels))

    @alias("get_channel_by_sn", "getChannelBySN")
    def get_channel_by_SN(self, SN):
        for node in self.root.goto("MassIntervalList"):
            if node.name == "mi":
                l = node.dict_list()
                if l["SN"]["utf16"] == SN:
                    name = l["assign"]["utf16"]
                    if not name:
                        name = l["desc"]["utf16"]
                    if not name:
                        name = "{:.2f}u".format(l["cmass"]["float"])
                    return name

    @alias("get_added_image_by_sn", "getAddedImageBySN")
    def get_added_image_by_SN(self, SN, raw=False):
        """
        New ITA fileformat save images with their respective serial number (SN).
        This function return the image for a given SN.

        Parameters
        ----------

        SN: Serial Number of the channel
        """
        node = self.root.goto(
            f"filterdata/TofCorrection/ImageStack/Reduced Data/Images/{SN}/SumImage/EDROff"
        )
        dat = node.decompress()
        data = struct.unpack(f"<{len(dat) // 4}I", dat)
        img = np.array(data, dtype=np.float64).reshape((self.sy, self.sx))
        if raw:
            return img
        channel = self.get_channel_by_SN(SN)
        return self.image(np.flipud(img), channel=channel)

    @alias("getAddedImage")
    def get_added_image(self, channel, **kargs):
        """
        Retrieve the numpy 2D array of a given channel ID for the sum of all scan (precomputed by iontof, but not shift-corrected)
        Note: It is preferable to use the pySPM.ITA.getAddedImageByMass or pySPM.ITA.getAddedImageByName
        """
        assert isinstance(channel, int)
        assert channel >= 0 and channel < self.Nimg
        c = self.root.goto(
            "filterdata/TofCorrection/ImageStack/Reduced Data/ImageStackScansAdded"
            "/Image[" + str(channel) + "]/ImageArray.Long"
        )
        D = zlib.decompress(c.value)
        V = np.array(
            struct.unpack("<" + str(self.sx * self.sy) + "I", D), dtype=np.float64
        ).reshape((self.sy, self.sx))
        return V

    @alias("fastGetImage")
    def fast_get_image(self, channel, scans, shifts=False, prog=False, **kargs):
        """
        Retieve a 2D numpy array corresponding to a given channel ID for given scan(s) and return their sum.

        Parameters
        ----------
        channel : int
            The channel ID
        scans: int, list of ints or 1D numpy array
            List of scans
        shifts : False or list of tuples
            List of the shift correction in pixels for ALL the scans ( not only the selected ones).
            If Flase not shift correction is performed
        prog : bool
            Display a progressbar ?

        Returns
        -------
        2D numpy array
            array data of the image
        """
        # Old parameter name compatibility
        if "Shifts" in kargs:
            shifts = kargs.pop("Shifts")

        Z = np.zeros((self.sy, self.sx))
        if prog:
            scans = PB(scans)

        im_root = self.root.goto(
            "filterdata/TofCorrection/ImageStack/Reduced Data/ImageStackScans/Image["
            + str(channel)
            + "]"
        )
        for scan in scans:
            c = im_root.goto("ImageArray.Long[" + str(scan) + "]")
            V = np.array(c.get_data("I"), dtype=np.float64).reshape((self.sy, self.sx))
            if shifts:
                r = [int(z) for z in shifts[scan]]
                V = np.roll(np.roll(V, -r[0], axis=1), -r[1], axis=0)
                rx = [max(0, -r[0]), self.sx - max(1, r[0])]
                ry = [max(0, -r[1]), self.sy - max(1, r[1])]
                Z[ry[0] : ry[1], rx[0] : rx[1]] += V[ry[0] : ry[1], rx[0] : rx[1]]
            else:
                Z += V
        return Z

    @alias("getImage")
    def get_image(
        self, channel, scan, shifts=None, shift_mode="roll", const=0, **kargs
    ):
        """
        getImage retrieve the image of a specific channel (ID) and a specific scan.

        Parameters
        ----------
        channel : int
            channel ID
        scan : int
            scan number (start with 0)
        Shifts : None or list of tuples
                None: No shift
                list of tuple in the form of where each tuple is in the form (dx,dy) for a given scan
        ShiftMode :  string
            roll : roll the data over. easy but non-physical
            const : replace missing values by a constant (given by argument const)
            NaN : the same as const but with const=numpy.NaN
        const : float
            if ShiftMode is 'const' then this parameter defines the constant used (default 0)
        """
        # Compatibility with old parameter names
        if "Shifts" in kargs:
            shifts = kargs.pop("Shifts")
        if "ShiftMode" in kargs:
            shift_mode = kargs.pop("ShiftMode")

        assert isinstance(channel, int)
        assert isinstance(scan, int)
        assert channel >= 0 and channel < self.Nimg
        assert scan >= 0 and scan < self.Nscan
        c = self.root.goto(
            "filterdata/TofCorrection/ImageStack/Reduced Data/ImageStackScans"
            "/Image[" + str(channel) + "]/ImageArray.Long[" + str(scan) + "]"
        )
        V = np.array(c.get_data(), dtype=np.float64).reshape((self.sy, self.sx))
        if shifts is not None:
            r = [int(z) for z in shifts[scan]]
            V = np.roll(np.roll(V, -r[0], axis=1), -r[1], axis=0)
            if shift_mode == "const" or shift_mode == "NaN":
                if shift_mode == "NaN":
                    const = np.nan
                if r[1] < 0:
                    V[: -r[1], :] = const
                elif r[1] > 0:
                    V[-r[1] :, :] = const
                if r[0] < 0:
                    V[:, : -r[0]] = const
                elif r[0] > 0:
                    V[:, -r[0] :] = const
        return V

    @alias("getOperation")
    def get_opertion(self, OpID):
        """
        Test function to retrieve the operations used in the Worksheet.
        """
        Nop = self.root.goto(
            "Presentation/Imaging Worksheet/Worksheet/OPERATIONS/OpCount"
        ).get_ulong()
        for i in range(Nop):
            blk = self.root.goto(
                f"Presentation/Imaging Worksheet/Worksheet/OPERATIONS/Operation[{i}]"
            )
            if blk.goto_item("OpID").get_ulong() == OpID:
                return blk
        return None

    @alias("showWorksheet")
    def show_worksheet(self, page=0):
        """
        In Dev. function to display the worksheet
        """
        import matplotlib as mpl

        from .utils import sp

        num_pages = self.root.goto(
            "Presentation/Imaging Worksheet/Worksheet/PAGES/COUNT"
        ).get_ulong()
        assert page < num_pages
        self.root.goto(
            f"Presentation/Imaging Worksheet/Worksheet/PAGES/Page[{page}]/ItemCount"
        ).get_ulong()
        sett = self.root.goto(
            f"Presentation/Imaging Worksheet/Worksheet/PAGES/Page[{page}]/SETTINGS"
        ).dict_list()
        sett["Xsize"]["ulong"]
        sett["Ysize"]["ulong"]
        items = self.root.goto(
            f"Presentation/Imaging Worksheet/Worksheet/PAGES/Page[{page}]/Items"
        ).get_data()
        ax = sp(len(items))
        IntV = {}
        for x in self.root.goto("MassIntervalList"):
            if x.name == "mi":
                d = x.dict_list()
                IntV[d["id"]["long"]] = d["desc"]["utf16"] + d["assign"]["utf16"]
        for i, it in enumerate(items):
            blk = self.get_operation(it)
            OPTYPE = blk.goto_item("OPTYPE").get_ulong()
            while OPTYPE != 3:
                OPTYPE = blk.goto_item("OPTYPE").get_ulong()
                if OPTYPE == 4:
                    blk = self.getOperation(blk.goto_item("ArgOpIDs").get_ulong())
                elif OPTYPE == 3:
                    palette = np.array(
                        blk.goto_item("BMP-Palette").get_data("B")
                    ).reshape((256, 4))
                    B, G, R = palette[:, 0], palette[:, 1], palette[:, 2]
                    dimx = blk.goto("Cache/IImage-Cache-DimX").get_ulong()
                    dimy = blk.goto("Cache/IImage-Cache-DimY").get_ulong()
                    img = np.array(
                        blk.goto("Cache/IImage-Cache-Intensities").get_data("d")
                    ).reshape((dimy, dimx))
                    RGB = np.hstack([R[:, None], G[:, None], B[:, None]]) / 256
                    cm = mpl.colors.ListedColormap(RGB)
                    ax[i].imshow(img, cmap=cm)

    def add_new_images(self, miblock, scans=None, added=None, prog=False, **kargs):
        # Compatibility with old parameter names
        if "Scans" in kargs:
            scans = kargs.pop("Scans")
        if "Added" in kargs:
            added = kargs.pop("Added")

        assert scans is not None or added is not None
        lvl = 3  # zlib encoding level
        sy, sx = self.size["pixels"]["y"], self.size["pixels"]["x"]
        SN = miblock.goto("SN").get_string()
        if added is None:
            added_img = np.zeros((sy, sx), dtype=np.uint32)
        miblock.goto("id").get_ulong()
        if scans is not None:
            N = self.root.goto(
                "filterdata/TofCorrection/ImageStack/Reduced Data/ImageStackScans/Image.NumberOfImages"
            ).get_ulong()
        AN = self.root.goto(
            "filterdata/TofCorrection/ImageStack/Reduced Data/ImageStackScansAdded/Image.NumberOfImages"
        ).get_ulong()
        self.root.edit_block(
            f"filterdata/TofCorrection/ImageStack/Reduced Data/ImageStackScansAdded/Image[{AN}]",
            "Image.MassIntervalSN",
            SN.encode("utf8"),
        )
        self.root.edit_block(
            f"filterdata/TofCorrection/ImageStack/Reduced Data/ImageStackScansAdded/Image[{AN}]",
            "Image.XSize",
            struct.pack("<I", sx),
        )
        self.root.edit_block(
            f"filterdata/TofCorrection/ImageStack/Reduced Data/ImageStackScansAdded/Image[{AN}]",
            "Image.YSize",
            struct.pack("<I", sy),
        )
        if scans is not None:
            RS = range(self.Nscan)
            if prog:
                RS = PB(RS)
            for i in RS:
                img = np.flipud(scans[i].astype(np.uint32, casting="unsafe"))
                data = zlib.compress(
                    struct.pack(f"<{sx * sy}I", *np.ravel(img)), level=lvl
                )
                self.root.edit_block(
                    f"filterdata/TofCorrection/ImageStack/Reduced Data/ImageStackScans/Image[{N}]",
                    "ImageArray.Long",
                    data,
                    id=i,
                    _type=128,
                )
                if added is None:
                    added_img += img

        added = added_img if added is None else np.flipud(added)
        data = zlib.compress(
            struct.pack(
                f"<{sx * sy}I", *np.ravel(added.astype(np.uint32, casting="unsafe"))
            ),
            level=lvl,
        )
        self.root.edit_block(
            f"filterdata/TofCorrection/ImageStack/Reduced Data/ImageStackScansAdded/Image[{AN}]",
            "ImageArray.Long",
            data,
            _type=128,
        )

        self.root.edit_block(
            f"filterdata/TofCorrection/ImageStack/Reduced Data/ImageStackScansAdded/Image[{AN}]",
            "Image.PulsesPerPixel",
            struct.pack("<I", self.spp * self.Nscan),
        )
        self.root.edit_block(
            f"filterdata/TofCorrection/ImageStack/Reduced Data/ImageStackScansAdded/Image[{AN}]",
            "Image.MaxCountsPerPixel",
            struct.pack("<I", int(np.max(added))),
        )
        self.root.edit_block(
            f"filterdata/TofCorrection/ImageStack/Reduced Data/ImageStackScansAdded/Image[{AN}]",
            "Image.MinCountsPerPixel",
            struct.pack("<I", int(np.min(added))),
        )
        self.root.edit_block(
            f"filterdata/TofCorrection/ImageStack/Reduced Data/ImageStackScansAdded/Image[{AN}]",
            "Image.TotalCountsDbl",
            struct.pack("<d", np.sum(added)),
        )
        self.root.edit_block(
            f"filterdata/TofCorrection/ImageStack/Reduced Data/ImageStackScansAdded/Image[{AN}]",
            "Image.TotalCounts",
            struct.pack("<I", int(np.sum(added))),
        )

        if scans is not None:
            self.root.edit_block(
                "filterdata/TofCorrection/ImageStack/Reduced Data/ImageStackScans",
                "Image.NumberOfImages",
                struct.pack("<I", N + 1),
            )
        self.root.edit_block(
            "filterdata/TofCorrection/ImageStack/Reduced Data/ImageStackScansAdded",
            "Image.NumberOfImages",
            struct.pack("<I", AN + 1),
        )
        self.Nimg += 1


@aliased
class ITA_collection(Collection):
    """
    ITA_collection is a super class containing a collection of tof-sims images.
    for details on Collection see pySPM.collection.Collection
    """

    def __init__(
        self,
        filename,
        channels1=None,
        channels2=None,
        name=None,
        mass=False,
        strict=False,
    ):
        """
        Opening a ToF-SIMS ITA file as an image collection

        Parameters
        ----------
        filename : string
            The filename
        channels1 : None or a list of names
        channels2 : None or a list of names
            channels1 and channels2 can be list of names or masses if mass=True
        name : string or None
            Name of the collection. If None, the basename of the filename is used (e.g. path/myfile.ita => name=myfile)
        mass : bool
            if True the channel lists are in mass and not names
        strict : bool
            Is the channel name strict? (see pySPM.ITA.getChannelsByName)

        Returns
        -------
        pySPM.ITA_collection class
        """
        self.ita = ITA(filename)
        self.filename = filename
        self.PCA = None
        if name is None:
            name = os.path.basename(filename)
        self.name = name
        Collection.__init__(
            self,
            sx=self.ita.fov,
            sy=self.ita.fov * self.ita.sy / self.ita.sx,
            unit="m",
            name=name,
        )
        self.msg = ""
        if channels1 is None:
            mass = True
            masses = self.ita.get_masses()
            channels1 = [x["cmass"] for x in masses if x["id"] > 1]
        CHS = [channels1]
        if channels2 is not None:
            CHS.append(channels2)
        for channels in CHS:
            if channels is channels2:
                strict = False
            if isinstance(channels, list):
                for x in channels:
                    if mass:
                        try:
                            I = self.ita.get_added_image_by_mass(x)
                            m = masses[2 + channels1.index(x)]
                            if m["assign"] != "":
                                self.add(I, m["assign"])
                            else:
                                self.add(I, f"{x:.2f}u")
                        except:
                            pass
                    else:
                        Z, ch = self.ita.get_added_image_by_name(x, strict)
                        self.msg += f"{x}\n"
                        for z in ch:
                            self.msg += "\t{name} ({desc}), mass: {lower:.2f} - {upper:.2f}\n".format(
                                desc=z["desc"],
                                name=z["assign"],
                                lower=z["lmass"],
                                upper=z["umass"],
                            )
                        self.add(Z, x)
            elif isinstance(channels, dict):
                for x in channels:
                    if mass:
                        self.add(self.ita.get_added_image_by_mass(channels[x]), x)
                    else:
                        Z, ch = self.ita.get_added_image_by_name(channels[x], strict)
                        self.msg += f"{x}\n"
                        for z in ch:
                            self.msg += "\t{name} ({desc}), mass: {lower:.2f} - {upper:.2f}\n".format(
                                desc=z["desc"],
                                name=z["assign"],
                                lower=z["lmass"],
                                upper=z["umass"],
                            )
                        self.add(Z, x)
            else:
                raise TypeError(
                    f"Channels should be a list or a dictionnary. Got {type(channels)}"
                )

    def __getitem__(self, key):
        """
        Retrieve the image of a given channel

        Example
        -------
        >>> A = pySPM.ITA_collection("myfile.ita")
        >>> A['Au-']
        <pySPM.SPM.SPM_image at 0x????????>
        """
        if key not in self.channels:
            return None
        return self.channels[key]

    @alias("runPCA")
    def run_pca(self, channels=None):
        """
        Perform a Principle Component Analysis (PCA) on the channels

        Parameters
        ----------
        channels : None or list of strings
            List of channels to use for the PCA. If None all channels will be used.
        """
        from .PCA import ITA_PCA

        if channels is None:
            channels = self.channels.keys()
        self.PCA = ITA_PCA(self, channels)

    @alias("showPCA")
    def show_pca(self, num=None, loadings=True, **kargs):
        """
        Run PCA if not already done and display the PCA images.

        Parameters
        ----------
        num : int or None
            The number of PC component to display. If None display all PAs
        **kargs : additional parameters
            passed to pySPM.PCA.showPCA

        Returns
        -------
        None
            Plot num PCA into a 1 x num subplots

        """
        if self.PCA is None:
            self.run_pca()
        self.PCA.show_pca(num=num, loadings=loadings, **kargs)

    def loadings(self, num=None, ax=None):
        """
        Return a pandas DataFrame with the num first loadings

        Parameters
        ----------
        num : int or None
            The number of PC to use. If None use all PCs

        Note
        ----
        The results can be used in combination with pySPM.PCA.hinton to create nice hinton plots
        >>> col = pySPM.ITA_collection("myfile.ita")
        >>> L = col.loadings(3)
        >>> col.PCA.hincton(matrix=L)
        Display a hinton plot with num lines representing the strength of each loading. Blue means negative loadings and Red means positive ones.
        The size of each square is proportional to the absolute value of each loading.
        """
        if self.PCA is None:
            self.run_pca()
        L = self.PCA.loadings() if num is None else self.PCA.loadings()[:num]
        if ax is not None:
            if ax is True:
                self.PCA.hinton(matrix=L)
            else:
                self.PCA.hinton(matrix=L, ax=ax)
        return L

    @deprecated("StitchCorrection")
    def stitch_correction(self, channel, stitches, gauss=0, debug=False):
        """
        When an image is created by stitching of several images (while moving the stage during the measurement) the resulting image can have several artifacts due to charging.
        The goal of this function is the try to suppress this stitching artifacts by givings a channel name which is known to be homogeneous everywhere

        Parameters
        ----------
        channel : string
            name of a channel with a known homogeneous yield (i.e. where the visible variation of the yield is only due to charging and not to a material density variation
        stitches : list or tuple of two ints
            stitches=(N,M) where N x M is the numer of images stitched
        gauss : float
            if >0 a gauss filter will be applied on the reference image
        debug : bool
            if True returns additionally to the new collection also the reference image

        Returns
        -------
        pySPM.ITA_collection
            A new collection with corrected data

        """
        import copy

        from scipy.ndimage.filters import gaussian_filter

        N = ITA_collection(self.filename, [], name=self.name)
        size = next(iter(self.channels.values())).pixels.shape
        S = np.zeros((int(size[0] / stitches[0]), int(size[1] / stitches[1])))
        sy, sx = S.shape
        for i in range(stitches[0]):
            for j in range(stitches[1]):
                S += self.channels[channel].pixels[
                    sy * i : sy * (i + 1), sx * j : sx * (j + 1)
                ]
        S[S == 0] = 1
        if gauss > 0:
            S = gaussian_filter(S, gauss)
        for x in self.channels:
            F = np.zeros(size)
            for i in range(stitches[0]):
                for j in range(stitches[1]):
                    F[sy * i : sy * (i + 1), sx * j : sx * (j + 1)] = (
                        self.channels[x].pixels[
                            sy * i : sy * (i + 1), sx * j : sx * (j + 1)
                        ]
                        / S
                    )
            new_channel = copy.deepcopy(self[x])
            new_channel.pixels = F
            N.add(new_channel, x)
        if debug:
            return N, S
        return N



# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

"""
Various helper function for plotting data with matplotlib
"""

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np

from .misc import alias


@alias("plot_mask")
def plotMask(ax, mask, color, **kargs):
    """
    Create an overlay of a given color where mask is True.
    The transparency and other parameters can be adjusted with **kargs. see help(plt.imshow)

    ax: matplotlib axis
    mask: An array with value True where the mask should be plotted
    color: color of the mask
    **kargs: dictionnary of arguments which can be passed to imshow. Useful is mainly: alpha
    """
    import copy

    m = np.ma.masked_array(mask, ~mask)
    palette = copy.copy(plt.cm.gray)
    palette.set_over(color, 1.0)
    ax.imshow(m, cmap=palette, vmin=0, vmax=0.5, **kargs)


def offset_coord(xy, offset=(0, 0), ax=None, fig=None, unit="px"):
    if ax is None:
        ax = plt.gca()

    if fig is None:
        fig = plt.gcf()

    tr = ax.transData
    tri = tr.inverted()

    if unit == "px":
        offset = np.array(offset)
    elif unit == "ax":
        offset = ax.transAxes.transform(offset)
    elif unit == "fig":
        offset = fig.transFigure.transform(offset)
    return tri.transform(tr.transform(xy) + offset)


def Xdist(
    ax,
    left,
    right,
    y,
    color="r",
    linestyle=":",
    fmt="{dist:.1f}{unit}",
    xtransf=lambda x: x,
    va="bottom",
    ha="center",
    offset=(0, 2),
    **kargs,
):
    """
    Show the distance between to x-values

    ax: matplotlib axis
    left: x-value of the left mark
    right: x-value of the right mark
    y: height of the label
    color: color of the lines / labels
    linestyle: linestyle of the vertical lines
    fmt: Format string to parse the distance
    xtransf: a function can be passed to display the distance in an other unit that the x axis.
        example: xtransf=lambda x: x*1e3 to display the distance in nm on an axis that uses micro-meters.
    va: vertical alignment of the distance label
    ha: horizontal alignment of the distance label
    offset: (x,y) offset of the label in pixels
    **kargs: additional arguments.
        arguments starting with amn_ will be sent to the annotation
        arguments starting with arr_ will be sent to the arrow
        for example arr_color='r', ann_color='b' will display a red arrow with the distance in blue
    """

    ax.axvline(left, color=color, linestyle=linestyle)
    ax.axvline(right, color=color, linestyle=linestyle)
    ann = {"va": va, "ha": ha, "color": color}
    ann.update({k[3:]: kargs[k] for k in kargs if k.startswith("an_")})
    ax.annotate(
        fmt.format(dist=xtransf(right - left), unit=kargs.get("unit", "")),
        ({"center": 0.5 * (left + right), "left": right, "right": left}[ann["ha"]], y),
        offset,
        textcoords="offset pixels",
        **ann,
    )
    arr = {"arrowstyle": "<->", "color": color}
    arr.update({k[4:]: kargs[k] for k in kargs if k.startswith("arr_")})
    ax.annotate("", (left, y), (right, y), arrowprops=arr)


def Ydist(
    ax,
    down,
    up,
    x,
    color="r",
    linestyle=":",
    fmt="{dist:.2f}{unit}",
    ytransf=lambda y: y,
    rotation=90,
    va="center",
    ha="right",
    offset=(-2, 0),
    **kargs,
):
    """
    Show the distance between to x-values

    ax: matplotlib axis
    down: y-value of the bottom mark
    up: y-value of the top mark
    x: gorizontal position of the label
    color: color of the lines / labels
    linestyle: linestyle of the vertical lines
    fmt: Format string to parse the distance
    xtransf: a function can be passed to display the distance in an other unit that the x axis.
        example: xtransf=lambda x: x*1e3 to display the distance in nm on an axis that uses micro-meters.
    va: vertical alignment of the distance label
    ha: horizontal alignment of the distance label
    offset: (x,y) offset of the label in pixels
    **kargs: additional arguments.
        arguments starting with amn_ will be sent to the annotation
        arguments starting with arr_ will be sent to the arrow
        for example arr_color='r', ann_color='b' will display a red arrow with the distance in blue
    """

    ax.axhline(down, color=color, linestyle=linestyle)
    ax.axhline(up, color=color, linestyle=linestyle)
    ann = {"va": va, "ha": ha, "color": color}
    ann.update({k[3:]: kargs[k] for k in kargs if k.startswith("an_")})
    ax.annotate(
        fmt.format(dist=ytransf(up - down), unit=kargs.get("unit", "")),
        (x, {"center": 0.5 * (down + up), "top": up, "bottom": down}[ann["va"]]),
        offset,
        textcoords="offset pixels",
        rotation=rotation,
        **ann,
    )
    arr = {"arrowstyle": "<->", "color": color}
    arr.update({k[4:]: kargs[k] for k in kargs if k.startswith("arr_")})
    ax.annotate("", (x, down), (x, up), arrowprops=arr)


@alias("DualPlot")
def dual_plot(ax, col1="C0", col2="C1"):
    """
    Create a dual axis from ax and tune the color of the left/right axes to col1, col2 resp.
    """
    axb = ax.twinx()
    axb.spines["left"].set_color(col1)
    axb.spines["right"].set_color(col2)
    ax.yaxis.label.set_color(col1)
    axb.yaxis.label.set_color(col2)
    ax.tick_params(axis="y", colors=col1)
    axb.tick_params(axis="y", colors=col2)
    return axb


def sp(M, N=1, W=21, ravel=True, fig=False):
    """
    Shortcut for creating subplots with max width = W (default 21,
        which seems to correspond to 100% of the width in jupyter).
    Height is calculated in order to have square subplots.
    The layout is given by M columns and N lines.
    If N is negative, abs(N) is interpreted as the number of elements and the number of line will be calculated automatically in function of M.
    """
    if N < 0:
        tot = -N
        N = 1 + (tot - 1) // M
    f, ax = plt.subplots(N, M, figsize=(W, N * W / M))
    if ravel:
        if fig:
            return np.ravel(ax), f
        return np.ravel(ax)
    if fig:
        return ax, f
    return ax


def get_rect(img, bottom, top, left, right, ax, color="r"):
    """
    return a sub-area on a 2D-array img and display the boundaries on a matplotlib axis ax
    """
    ax.axhline(bottom, color=color)
    ax.axhline(top, color=color)
    ax.axvline(left, color=color)
    ax.axvline(right, color=color)
    return img[bottom:top, left:right]


def sublegend(*ax, **kargs):
    labels = kargs.get("labels", None)
    color = kargs.get("color", "white")
    margin = kargs.get("margin", 9)
    titles = kargs.get("titles", None)
    fontsize = kargs.get("fontsize", 14)
    """
    ax: list of axes
    labels: If None, the will be a, b, c, d, ...
    color: background color of the rectangle
    margin: margin in pixel from the axis border
    titles: set to False to remove all titles. (Useful to keep the set_title info in the code to remember what is plotted)
    fontsize: The font size of the labels
    """
    props = {"boxstyle": "round", "facecolor": color, "alpha": 1}
    if not hasattr(margin, "__getitem__") and not hasattr(margin, "__iter__"):
        margin = (margin, margin)

    if labels is None:
        labels = [chr(ord("a") + i) for i, _ in enumerate(np.ravel(ax))]
    for i, a in enumerate(np.ravel(ax)):
        if titles is False:
            a.set_title("")
        a.annotate(
            labels[i],
            (0, 1),
            xytext=(margin[0], -margin[1]),
            fontsize=fontsize,
            verticalalignment="top",
            bbox=props,
            xycoords="axes fraction",
            textcoords="offset pixels",
        )


def formula(x):
    import re

    x = re.sub("_([0-9]+)", r"$_{\1}$", x)
    x = re.sub(r"\^([0-9\+\-]+)", r"$^{\1}$", x)
    return x


def _points_in_bbox(x, y, bbox):
    x_in = np.logical_and(x >= bbox.xmin, x <= bbox.xmax)
    y_in = np.logical_and(y >= bbox.ymin, y <= bbox.ymax)
    mask = x_in & y_in
    return np.any(mask), mask


def _bbox_overlap(bbox1, bbox2):
    x_overlap = (bbox1.xmax >= bbox2.xmin) and (bbox1.xmin <= bbox2.xmax)
    y_overlap = (bbox1.ymax >= bbox2.ymin) and (bbox1.ymin <= bbox2.ymax)
    return x_overlap and y_overlap


_cached_points = {}
_cached_pointsd = {}


def _overlap(ax, obj, objs, debug=False):
    global _cached_points
    fig = ax.get_figure()
    renderer = fig.canvas.get_renderer()
    tr = ax.transData
    #    tri = tr.inverted()
    r = obj.get_window_extent(renderer)  # got object bbox
    for o in objs:
        if type(o) is mpl.lines.Line2D:
            if o in _cached_points:
                xy = _cached_points[o]
                xyd = _cached_pointsd[o]
            else:
                xyd = np.vstack(o.get_data()).T  # retriev data in data coordinates
                xy = tr.transform(xyd)  # retrieve data in pixel coordinates
                _cached_points[o] = xy
                _cached_pointsd[o] = xyd
            isin, ins = _points_in_bbox(xy[:, 0], xy[:, 1], r)
            if isin:
                if debug:
                    ax.plot(xyd[:, 0][ins], xyd[:, 1][ins], "rx")
                return o
        else:
            ro = o.get_window_extent(renderer)
            if _bbox_overlap(r, ro):
                return o
    return False


def color_frame(ax, color, sx=0.01, sy=None, lw=5, **kargs):
    if sy is None:
        sy = sx
    import matplotlib as mpl

    fig = plt.gcf()
    fig.patches.append(
        mpl.patches.Rectangle(
            (-sx, -sy),
            1 + 2 * sx,
            1 + 2 * sy,
            transform=ax.transAxes,
            ec=color,
            fill=False,
            lw=lw,
            zorder=0,
            **kargs,
        )
    )


def _get_yextend_of_curve(ax, bbox, curve):
    tri = ax.transData.inverted()
    rd = bbox.transformed(tri)  # bbox in data coordinates
    x, y = curve.get_data()
    mask = (x >= rd.xmin) * (x <= rd.xmax)
    if len(y[mask]) == 0:
        return (0, 0)
    return (np.min(y[mask]), np.max(y[mask]))


def put_Xlabels(
    ax, pos, labels, colors="rgb", debug=False, save=False, bbox=False, **kargs
):
    fig = ax.get_figure()
    renderer = fig.canvas.get_renderer()
    P = list(zip(pos, labels))
    P.sort(key=lambda x: x[0])  # sort labels by ascending x
    labs = []
    ax.draw(renderer)  # make sure to draw the new object
    ylim = ax.get_ylim()
    objs = [o for o in ax.get_children() if type(o) not in [mpl.patches.Rectangle]]
    coli = [None, None]
    for i, (x, lab) in enumerate(P):
        col = colors[i % len(colors)]
        pos = offset_coord((x, ylim[0]), (3, 0), ax=ax)
        labs.append(
            ax.text(pos[0], pos[1], lab, va="bottom", ha="left", rotation=90, color=col)
        )
        ax.axvline(
            x,
            color=col,
            linestyle=kargs.get("linestyle", "-"),
            alpha=kargs.get("line_alpha", 0.5),
        )
        ax.draw(renderer)  # make sure to draw the new object
        r = labs[-1].get_window_extent(renderer)  # get the Bbox of the last label
        ov = _overlap(ax, labs[-1], objs + labs[:-1])
        if save:
            if ov:
                r1 = r.transformed(ax.transData.inverted())
                r2 = ov.get_window_extent(renderer).transformed(ax.transData.inverted())
                print(repr(ov), r2.ymin, r2.ymax)
                coli[0] = mpl.patches.Rectangle(
                    (r1.xmin, r1.ymin),
                    r1.xmax - r1.xmin,
                    r1.ymax - r1.ymin,
                    ec="b",
                    fill=False,
                )
                coli[1] = mpl.patches.Rectangle(
                    (r2.xmin, r2.ymin),
                    r2.xmax - r2.xmin,
                    r2.ymax - r2.ymin,
                    ec="r",
                    fill=False,
                )

                ax.add_patch(coli[0])
                ax.add_patch(coli[1])
            fig.savefig(f"put_Xlabels_{i:03d}_000.png")
            if ov:
                coli[0].remove()
                coli[1].remove()
        last_ov = ov
        y_offset = 0
        it = 0
        while ov:
            it += 1
            if debug:
                print(f'Label "{labs[-1].get_text()}" overlap with {ov!r}')
            if type(ov) is mpl.lines.Line2D:
                new_y = _get_yextend_of_curve(ax, r, ov)[1]
            else:
                rov = ov.get_window_extent(renderer).transformed(
                    ax.transData.inverted()
                )
                new_y = rov.ymax
            if ov == last_ov:
                y_offset += kargs.get("offset_step", 5)
                if y_offset > kargs.get("offset_step", 5) * kargs.get(
                    "max_iteration", 6
                ):
                    break
            else:
                y_offset = kargs.get("offset_step", 5)
            last_ov = ov
            new_xy_px = ax.transData.transform((x, new_y)) + np.array([3, y_offset])
            new_xy = ax.transData.inverted().transform(
                new_xy_px
            )  # Offset the new_y of 5 pixels
            labs[-1].set_position(new_xy)
            ax.draw(renderer)
            ov = _overlap(ax, labs[-1], objs + labs[:-1], debug=debug)
            if save:
                ax.draw(renderer)
                fig.savefig(f"put_Xlabels_{i:03d}_{it:03d}.png")
                if ov:
                    r2 = ov.get_window_extent(
                        renderer
                    )  # get the Bbox of the last label
                    mpl.patches.draw_bbox(r, renderer, color="r")
                    mpl.patches.draw_bbox(r2, renderer, color="r")
    if bbox:
        ax.draw(renderer)
        for o in objs:
            if type(o) is not mpl.lines.Line2D:
                r = o.get_window_extent(renderer).transformed(ax.transData.inverted())
                ax.add_patch(
                    mpl.patches.Rectangle(
                        (r.xmin, r.ymin),
                        r.xmax - r.xmin,
                        r.ymax - r.ymin,
                        ec="r",
                        fill=False,
                    )
                )

        for l in labs:
            r = l.get_window_extent(renderer).transformed(ax.transData.inverted())
            ax.add_patch(
                mpl.patches.Rectangle(
                    (r.xmin, r.ymin),
                    r.xmax - r.xmin,
                    r.ymax - r.ymin,
                    ec="b",
                    fill=False,
                )
            )


def stdplot(x, y, Nsig=2, ax=None, color="b", axis=1, **kargs):
    """
    Plot the mean / standard-deviation for a signal

    x: 1D numpy.ndarrayx
        x values
    y: 2D numpy.ndarray
        y values. One of the two axis is used for the statistics
    Nsig: int
        The number of standard-deviations to display
    axis: int
        which axis is used to calculate mean/std
    ax: matplotlib axis
        maplotlib axis to plot in
    color: string
        color of the line
    **kargs: arguments passed to plot

    """
    if ax is None:
        ax = plt.gca()
    dy = np.std(y, axis=axis)
    mean = np.mean(y, axis=axis)
    ax.plot(x, mean, color=color, **kargs)
    for i in range(1, Nsig + 1):
        ax.fill_between(x, mean - i * dy, mean + i * dy, color=color, alpha=0.2)


def pixel2img(xy, ax=None):
    """
    Plot create by SPM.show can be either in pixels or real unit.
    This function return the input in case the plot is in pixels
    or calculate and return the real unit coordinate of the corresponding pixel.
    """
    import matplotlib.pyplot as plt

    if ax is None:
        ax = plt.gca()

    if hasattr(ax, "isPixel") and not ax.isPixel:
        pixel_shape = ax.images[0].get_array().shape
        extent = ax.images[0].get_extent()
        eW = extent[1] - extent[0]
        eH = extent[3] - extent[2]
        return (xy[0] * eW / pixel_shape[1], xy[1] * eH / pixel_shape[0])
    return xy



import sys

from PyQt5.QtCore import QTimer
from PyQt5.QtWidgets import QApplication, QMainWindow

from pySPM.tools.timer_display import Ui_ToF_Timer
from pySPM.tools.win32_helper import findWindow, getText


class GUI_Timer(QMainWindow):
    def __init__(self):
        super(QMainWindow, self).__init__()
        self.ui = Ui_ToF_Timer()
        self.ui.setupUi(self)
        self.timer = QTimer(self)
        self.timer.timeout.connect(self.update)
        self.timer.start(1000)

    def update(self):
        ok = False
        A = findWindow("Measurement Progress")
        if len(A) > 0:
            B = findWindow(C="Edit", parent=A[0])
            if len(B) > 0:
                AnalTime = int(getText(B[2]).replace(",", ""))
                TotScans = int(getText(B[1]).replace(",", ""))
                Scans = int(getText(B[0]).replace(",", ""))
                self.ui.label_2.setText(f"Scans: {Scans} / {TotScans}")
                self.ui.label_3.setText(f"Analysis Time: {AnalTime} s")
                self.ui.progressBar.setValue(Scans)
                self.ui.progressBar.setMaximum(TotScans)
                ok = True
        if not ok:
            self.ui.label.setText(
                "Remaining time: Unavailable (measurement not in progress?)"
            )
            return
        if Scans > 0:
            r = AnalTime * (TotScans - Scans) / Scans
            h = int(r // 3600)
            m = int((r - h * 3600) // 60)
            s = int(r - h * 3600 - m * 60)
            self.ui.label.setText(f"Remaining time: {h:02d}:{m:02d}:{s:02d}")
        else:
            self.ui.label.setText("Remaining time: Unknown")


def main():
    app = QApplication(sys.argv)
    window = GUI_Timer()
    window.show()
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()



# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>

"""
This module handle sxm file format used by Nanonis instruments
"""

import os
import struct

import numpy as np

from .SPM import SPM_image

# -- coding: utf-8 --
# Copyright 2018 Olivier Scholder <o.scholder@gmail.com>


class SXM:
    def __init__(self, filename):
        self.filename = filename
        assert os.path.exists(self.filename)
        self.f = open(self.filename, "rb")
        l = ""
        key = ""
        self.header = {}
        while l != b":SCANIT_END:":
            l = self.f.readline().rstrip()
            if l[:1] == b":":
                key = l.split(b":")[1].decode("ascii")
                self.header[key] = []
            else:
                if l:  # remove empty lines
                    try:
                        self.header[key].append(l.decode("ascii").split())
                    except UnicodeDecodeError:
                        self.header[key].append(l.decode("unicode_escape").split())

        while self.f.read(1) != b"\x1a":
            pass
        assert self.f.read(1) == b"\x04"
        assert self.header["SCANIT_TYPE"][0][0] in ["FLOAT", "INT", "UINT", "DOUBLE"]
        self.data_offset = self.f.tell()
        self.f.close()
        self.size = {
            "pixels": {
                "x": int(self.header["SCAN_PIXELS"][0][0]),
                "y": int(self.header["SCAN_PIXELS"][0][1]),
            },
            "real": {
                "x": float(self.header["SCAN_RANGE"][0][0]),
                "y": float(self.header["SCAN_RANGE"][0][1]),
                "unit": "m",
            },
        }

    def list_channels(self):
        print("Channels")
        print("========")
        h = self.header["DATA_INFO"][0]
        i = h.index("Name")
        for z in self.header["DATA_INFO"][1:]:
            print("  - " + z[i])

    def get_channel(self, name, direction="forward", corr=None):
        chID = 0
        zscale = ""
        for x in self.header["DATA_INFO"][1:]:
            if x[1] == name:
                if x[3] == "both" and direction == "backward":
                    chID += 1
                if x[3] == "both" or direction == x[3]:
                    break
                return None
            elif x[3] == "both":
                chID += 2
            else:
                chID += 1

        size = self.size["pixels"]["x"] * self.size["pixels"]["y"]
        self.f = open(self.filename, "rb")
        self.f.seek(self.data_offset + chID * size * 4)
        data = np.array(
            struct.unpack(
                "<>"[self.header["SCANIT_TYPE"][0][1] == "MSBFIRST"]
                + str(size)
                + {"FLOAT": "f", "INT": "i", "UINT": "I", "DOUBLE": "d"}[
                    self.header["SCANIT_TYPE"][0][0]
                ],
                self.f.read(4 * size),
            )
        ).reshape((self.size["pixels"]["y"], self.size["pixels"]["x"]))
        self.f.close()
        return SPM_image(
            channel=name,
            BIN=data,
            real=self.size["real"],
            _type="Nanonis SXM",
            zscale=zscale,
            corr=corr,
        )



#!python
import os
import sys

from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg as FigureCanvas
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT as NavigationToolbar
from matplotlib.figure import Figure
from PyQt5.QtWidgets import (
    QAction,
    QApplication,
    QComboBox,
    QFileDialog,
    QMainWindow,
    QVBoxLayout,
    QWidget,
)

from pySPM import ITM


def DualPlot(ax, col1="C0", col2="C1"):
    axb = ax.twinx()
    axb.spines["left"].set_color(col1)
    axb.spines["right"].set_color(col2)
    ax.yaxis.label.set_color(col1)
    axb.yaxis.label.set_color(col2)
    ax.tick_params(axis="y", colors=col1)
    axb.tick_params(axis="y", colors=col2)
    return axb


class Plotter(QMainWindow):
    def plot(self, ev=None, scans=3):
        if self.cwd is not None:
            self.figure.clf()
            ax = self.figure.add_subplot(1, 1, 1)

            filename = os.path.join(self.cwd, self.fileDrop.currentText())
            I = ITM(filename)
            I.show_stability(ax)
            del I
            self.canvas.draw()

    def open(self, dirs=None):
        if dirs is None:
            dirs = QFileDialog.getExistingDirectory(self)
        if dirs:
            self.cwd = dirs
            self.fileDrop.clear()
            for x in os.listdir(dirs):
                if x.endswith((".itm", ".ITM")):
                    self.fileDrop.addItem(x)

    def __init__(self):
        QWidget.__init__(self)
        self.cwd = None
        self.figure = Figure()
        self.canvas = FigureCanvas(self.figure)
        self.toolbar = NavigationToolbar(self.canvas, self)
        self.markers = None
        self.draggable = None
        self.img = None
        self.msize = 6

        openAction = QAction("&Open", self)
        openAction.setShortcut("Ctrl+O")
        openAction.setStatusTip("Open folder")
        openAction.triggered.connect(self.open)

        menubar = self.menuBar()
        fileMenu = menubar.addMenu("&File")
        fileMenu.addAction(openAction)

        self.fileDrop = QComboBox()

        layout = QVBoxLayout()
        layout.addWidget(self.fileDrop)
        layout.addWidget(self.toolbar)
        layout.addWidget(self.canvas)

        window = QWidget()
        window.setLayout(layout)
        self.setCentralWidget(window)

        self.fileDrop.currentIndexChanged.connect(self.plot)

        self.show()


def main():
    app = QApplication(sys.argv)
    a = Plotter()
    a.open()
    sys.exit(app.exec_())


if __name__ == "__main__":
    main()
